# encoding: utf-8

'''
@author: vmmhypervisor

:required:
   https://pypi.python.org/pypi/XlsxWriter
'''

import six

from golib.logger import logger
import golib.common as common
import golib.writer as writer
import golib.data_property as dataprop

IMPORT_ERROR_MESSAGE = "import error: xlsxwriter package not found"

try:
    import xlsxwriter
    LOAD_XLSXWRITER = True

    logger.debug("import success: xlsxwriter")
except ImportError, e:
    LOAD_XLSXWRITER = False
    logger.warn(IMPORT_ERROR_MESSAGE)


def xl_rowcol_to_cell(row, col):
    return xlsxwriter.utility.xl_rowcol_to_cell(row, col)


class ExcelWorkbook(common.BaseObject):
    EXTENSION = ".xlsx"

    @property
    def workbook(self):
        return self.__workbook

    @property
    def file_path(self):
        return self.__file_path

    def __init__(self):
        super(ExcelWorkbook, self).__init__()

        self.__workbook = None
        self.__file_path = None

    def __del__(self):
        self.close()

    def open(self, file_path):
        if not LOAD_XLSXWRITER:
            raise ImportError(IMPORT_ERROR_MESSAGE)

        common.verifyVarType(common.isString, file_path)

        logger.debug("open workbook: %s" % (file_path))

        self.__file_path = file_path
        self.__workbook = xlsxwriter.Workbook(file_path)

    def close(self):
        if self.workbook is None:
            return

        logger.debug("close workbook: %s" % (self.file_path))
        self.__workbook.close()

        self.__init__()


def getSheetTable(name_list):
    """
    name_listから { <name>: <sheet name>} を返す。
    """

    sheet_table = {"toc": "toc"}
    sheet_idx = 0

    digit = common.getSignificantDigit(len(name_list))
    sheet_format = "%0" + str(digit) + "d"
    #sheet_format	= "s%0" + str(digit) + "d"

    for name in name_list:
        sheet_table[name] = sheet_format % (sheet_idx)
        sheet_idx += 1

    common.debugDict(sheet_table, locals())

    return sheet_table


class ExcelWriter(writer.NullWriter):

    class FormatName:
        HEADER = "header"
        CELL = "cell"
        NAN = "nan"

    class Default:
        FONT_NAME = "MS Gothic"
        FONT_SIZE = 11

        CELL_FORMAT = {
            "font_name" 	: FONT_NAME,
            "font_size" 	: FONT_SIZE,
            "align"			: "top",
            "text_wrap"		: True,
            "top"			: 1,
            "left"			: 1,
            "bottom"		: 1,
            "right"			: 1,
        }
        HEADER_FORMAT = {
            "font_name" 	: FONT_NAME,
            "font_size" 	: FONT_SIZE,
            "bg_color"		: "#DFDFFF",
            "bold"			: True,
            "left"			: 1,
            "right"			: 1,
            "rotation"		: 90,
        }
        NAN_FORMAT = {
            "font_name"		: FONT_NAME,
            "font_size"		: FONT_SIZE,
            "font_color"	: "silver",
            "top"			: 1,
            "left"			: 1,
            "bottom"		: 1,
            "right"			: 1,
        }

    @property
    def workbook(self):
        return self.__workbook

    def __init__(self):
        super(ExcelWriter, self).__init__()

        self.default_format = self.Default.CELL_FORMAT
        self.dict_format = {
            self.FormatName.CELL	: self.Default.CELL_FORMAT,
            self.FormatName.HEADER	: self.Default.HEADER_FORMAT,
            self.FormatName.NAN		: self.Default.NAN_FORMAT,
        }

        self.dict_ColumnLevel_PatternList = {}
        self.dict_SparkCol_Range = {}

        self.stream = None
        self.setWorkbook(None)

        self.__first_header_row = 2
        self.__last_header_row = self.first_header_row + 1

        self.__first_data_row = self.last_header_row + 1
        self.__first_data_col = 0

        self.__last_data_row = None
        self.__last_data_col = None

        self.__dict_sheet = {}

    @property
    def first_header_row(self):
        return self.__first_header_row

    @property
    def last_header_row(self):
        return self.__last_header_row

    @property
    def first_data_row(self):
        return self.__first_data_row

    @property
    def last_data_row(self):
        return self.__last_data_row

    @property
    def first_data_col(self):
        return self.__first_data_col

    def setFirstDataCol(self, value):
        self.__first_data_col = value

    @property
    def last_data_col(self):
        return self.__last_data_col

    def setWorkbook(self, workbook):
        """
        argument:
                ExcelWorkbook
        """

        self.__workbook = workbook

    def add_format(self, dict_property):
        return self.workbook.workbook.add_format(dict_property)

    def make_worksheet(self, sheet_name):
        if sheet_name in self.__dict_sheet:
            self.stream = self.__dict_sheet.get(sheet_name)
            return

        self.__dict_sheet[sheet_name] = self.workbook.workbook.add_worksheet(
            sheet_name)
        self.stream = self.__dict_sheet.get(sheet_name)

    def writeChapter(self, text):
        props = {
            "font_size" 	: 14,
            "bold"			: True,
            "underline"		: 1,
        }

        self.stream.write(0, 0, text, self.add_format(props))

    def __get_number_property(self, col_prop_list):
        dict_Col_NumProps = {}
        for col, col_prop in enumerate(col_prop_list):
            num_props = {}
            if common.isInteger(col_prop.minmax_decimal_places.max_value):
                float_digit = col_prop.minmax_decimal_places.max_value
                if float_digit > 0:
                    num_props = {"num_format": "0.%s" % ("0" * float_digit)}

            dict_Col_NumProps[col] = num_props

        return dict_Col_NumProps

    def _setCellWidthEx(self, col_prop_list, font_size):
        MAX_CELL_WIDTH = 60

        for col_idx, col_prop in enumerate(col_prop_list):
            width = min(col_prop.padding_len, MAX_CELL_WIDTH) * \
                (font_size / 10.0) + 1
            self.stream.set_column(col_idx, col_idx, width=width)

    def printListList(self, header_list, data_matrix, comment_matrix=None):
        common.debugDict(self.dict_format, locals())

        header_format_props = self.dict_format.get(
            self.FormatName.HEADER, self.default_format)
        header_format = self.add_format(header_format_props)
        nan_format_props = self.dict_format.get(
            self.FormatName.NAN, self.default_format)
        cell_format_props = self.dict_format.get(
            self.FormatName.CELL, self.default_format)

        col_prop_list = dataprop.PropertyExtractor.extractColumnPropertyList(
            [], data_matrix)
        # header_list, data_matrix)
        value_prop_matrix = dataprop.PropertyExtractor.extractDataPropertyMatrix(
            data_matrix)

        # set cell size ---
        self._setCellWidthEx(
            col_prop_list, cell_format_props.get("font_size"))

        # output header ---
        if common.isNotEmptyListOrTuple(header_list):
            self.stream.write_row(
                row=self.first_header_row, col=0,
                data=header_list, cell_format=header_format)
            self.stream.write_row(
                row=self.last_header_row, col=0,
                data=[""] * len(header_list), cell_format=header_format)

        # output data ---
        dict_Col_NumProps = self.__get_number_property(col_prop_list)
        self.dict_col_cell_format = {}

        for row, value_prop_list in enumerate(value_prop_matrix):
            sheet_row = self.first_data_row + row

            for col, prop in enumerate(value_prop_list):
                # write data ---
                if common.isEmptyString(prop.data):
                    continue

                base_props = dict(cell_format_props)
                key = "%d_%d" % (col, prop.typecode)

                if prop.typecode in [dataprop.Typecode.INT, dataprop.Typecode.FLOAT]:
                    num_props = dict_Col_NumProps.get(col, {})
                    base_props.update(num_props)

                    cell_format = self.dict_col_cell_format.get(key)
                    if cell_format is None:
                        # cache miss
                        cell_format = self.add_format(base_props)
                        self.dict_col_cell_format[key] = cell_format

                    try:
                        self.stream.write_number(
                            sheet_row, col, float(prop.data), cell_format)
                    except TypeError, e:
                        logger.error(prop.data)
                else:
                    if prop.data is None:
                        base_props = dict(nan_format_props)

                    cell_format = self.dict_col_cell_format.get(key)
                    if cell_format is None:
                        # cache miss
                        cell_format = self.add_format(base_props)
                        self.dict_col_cell_format[key] = cell_format

                    self.stream.write(
                        sheet_row, col, prop.data, cell_format)

                # write comment ---
                if common.isEmptyListOrTuple(comment_matrix):
                    continue

                try:
                    comment = comment_matrix[row][col]
                    if common.isEmptyString(comment):
                        continue
                    self.stream.write_comment(sheet_row, col, comment)
                except KeyError, e:
                    logger.exception(e)

        self.__last_data_col = len(value_prop_list)

        # set group ---
        import re
        for level, pattern_list in six.iteritems(self.dict_ColumnLevel_PatternList):
            args = {
                "level"		: level,
                "collapsed"	: True,
                "hidden"	: True,
            }

            level_color_table = {
                1	: "#CFCFFF",
                2	: "#BFBFFF",
            }

            for col, header in enumerate(header_list):
                props = dict(header_format_props)

                for pattern in pattern_list:
                    if re.search(pattern, header) is None:
                        continue

                    self.stream.set_column(col, col, options=args)

                    props["bg_color"] = level_color_table.get(level)
                    cell_format = self.add_format(props)
                    self.stream.write_string(
                        self.first_header_row, col, header, cell_format)
                    self.stream.write_string(
                        self.last_header_row, col, "", cell_format)

                    break

        # output spark line  ---
        for col, col_range in six.iteritems(self.dict_SparkCol_Range):
            for row in range(len(data_matrix)):
                sparkline_range = "%s!%s:%s" % (
                    self.stream.get_name(),
                    xlsxwriter.utility.xl_rowcol_to_cell(
                        self.first_data_row + row, col_range[0]),
                    xlsxwriter.utility.xl_rowcol_to_cell(
                        self.first_data_row + row, col_range[1]))
                parameter = {
                    "range"	: sparkline_range,
                    "type"	: "column",
                    "style"	: 6,
                    "show_hidden"	: True,
                }
                self.stream.add_sparkline(
                    self.first_data_row + row, col, parameter)

        def getLastColumn(header_list, data_matrix):
            if common.isNotEmptyListOrTuple(header_list):
                return len(header_list)

            if common.isNotEmptyListOrTuple(data_matrix):
                return len(data_matrix[0])

            raise ValueError()

        self.__last_data_row = self.first_data_row + len(data_matrix)
        self.__last_data_col = getLastColumn(header_list, data_matrix)
        self.stream.autofilter(
            self.last_header_row, self.first_data_col,
            self.last_data_row, self.last_data_col)
        self.stream.freeze_panes(self.first_data_row, self.first_data_col)


def makeExcelWriter(file_name, sheet_name):
    wb_path = file_name + ".xlsx"
    wb = ExcelWorkbook()
    wb.open(wb_path)

    writer = ExcelWriter()
    writer.setWorkbook(wb)
    writer.make_worksheet(sheet_name)

    logger.info("output: " + wb_path)

    return writer
