# encoding: utf-8

'''
@author: vmmhypervisor

:required:
   https://pypi.python.org/pypi/XlsxWriter
'''

from __future__ import with_statement
import sys
import os.path
import re

import golib.common as common
from golib.logger import logger


class FileType:
    FILE = 1
    DIRECTORY = 2
    LINK = 3


class InvalidFilePathError(Exception):
    pass


class FileNotFoundError(Exception):
    pass


class EmptyFileError(Exception):
    pass


RECOVERABLE_ERROR_LIST = (
    InvalidFilePathError, FileNotFoundError, EmptyFileError)


class FileTypeChecker:

    class FileType:
        TEXT = "ASCII text"
        BINARY = "data"

    __re_text = re.compile(FileType.TEXT)

    @classmethod
    def get_file_type(cls, file_path):
        import magic

        return magic.from_file(file_path)

    @classmethod
    def is_text_file(cls, file_path):
        return cls.__re_text.search(cls.get_file_type(file_path)) is not None


class FileManager:
    __dry_run = False

    @classmethod
    def initialize(cls, dry_run):
        cls.__dry_run = dry_run

    @classmethod
    def touch(cls, path):
        logger.debug("touch file: " + path)

        if cls.__dry_run:
            return

        cls.makeDirectory(os.path.dirname(path))

        with open(path, "a") as _fp:
            pass

    @classmethod
    def makeDirectory(cls, path, force=False):
        common.verifyVarType(common.isString, path)

        try:
            check_file_existence(path)
        except (FileNotFoundError, EmptyFileError):
            pass
        except InvalidFilePathError:
            return False
        else:
            logger.debug("already exists: " + path)
            return False

        logger.debug("make directory: " + path)
        if not cls.__dry_run or force:
            os.makedirs(path)

        return True

    @classmethod
    def copyFile(cls, src_path, dst_path):
        import shutil

        try:
            check_file_existence(src_path)
        except FileNotFoundError, e:
            logger.debug(e)
            return False
        except EmptyFileError:
            pass

        dst_file_type = None
        try:
            dst_file_type = check_file_existence(dst_path)
        except:
            pass
        if dst_file_type == FileType.DIRECTORY:
            return False

        logger.debug("copy: %s -> %s" % (src_path, dst_path))
        if cls.__dry_run:
            return True

        try:
            shutil.copyfile(src_path, dst_path)
        except (shutil.Error, IOError), e:
            logger.debug("skip copy: " + str(e))

        return True

    @classmethod
    def moveFile(cls, src_path, dst_path):
        import shutil

        try:
            check_file_existence(src_path)
        except FileNotFoundError, e:
            logger.debug(e)
            return False
        except EmptyFileError:
            pass

        common.verifyVarType(common.isString, dst_path)

        if os.path.realpath(src_path) == os.path.realpath(dst_path):
            logger.debug("%s and %s are the same file" % (
                src_path, dst_path))
            return True

        logger.info("move: %s -> %s" % (src_path, dst_path))
        if not cls.__dry_run:
            shutil.move(src_path, dst_path)

        return True

    @classmethod
    def chmod(cls, path, permission_text):
        """
        permission_text: "ls -l" style permission: e.g. -rw-r--r--
        """

        try:
            check_file_existence(path)
        except FileNotFoundError, e:
            logger.debug(e)
            return False
        except EmptyFileError:
            pass

        logger.debug("chmod %s %s" % (path, permission_text))

        os.chmod(path, parseLsPermissionText(permission_text))

    @classmethod
    def renameFile(cls, src_path, dst_path):
        if src_path == dst_path:
            logger.debug("no need to rename: src_path == dst_path")
            return True

        if common.isEmptyString(src_path):
            logger.error("empty source path")
            return False

        if not os.path.exists(src_path):
            if not cls.__dry_run:
                logger.debug("directory not found: %s" % (src_path))
            return False

        if common.isEmptyString(dst_path):
            logger.error("empty destination path")
            return False

        if os.path.isdir(dst_path):
            if os.path.exists(dst_path):
                logger.error("%s already exists" % (dst_path))
                return False

        logger.info("rename: %s -> %s" % (src_path, dst_path))
        if not cls.__dry_run:
            os.rename(src_path, dst_path)

        return True

    @classmethod
    def removeDirectory(cls, path):
        if common.isEmptyString(path):
            return False

        if not os.path.isdir(path):
            return False

        logger.debug("remove directory: " + path)
        if not cls.__dry_run:
            try:
                import shutil

                shutil.rmtree(path, False)
            except (ImportError, IOError), e:
                logger.exception(e)
                return False

        return True

    @classmethod
    def removeFile(cls, path):
        if common.isEmptyString(path):
            logger.warn("null path")
            return False

        if not os.path.lexists(path):
            logger.debug("file not found: %s" % (path))
            return False

        if os.path.isfile(path) or os.path.islink(path):
            if os.path.isfile(path):
                logger.debug("remove file: %s" % (path))
            elif os.path.islink(path):
                logger.debug("remove link: %s" % (path))

            if not cls.__dry_run:
                try:
                    os.remove(path)
                except Exception, e:
                    logger.exception(e)
                    return False
            return True
        elif os.path.isdir(path):
            return cls.removeDirectory(path)
        else:
            logger.error("unknown file type: %s" % (path))
            return False

    @classmethod
    def removeMatchFileInDir(cls, search_dir_path, re_target_list):
        logger.debug("remove matched file: search-dir=%s, re=%s" % (
            search_dir_path, str(re_target_list)))

        dict_Result_PathList = {}

        if common.isEmptyString(search_dir_path) or not os.path.isdir(search_dir_path):
            logger.debug("directory not found: " + str(search_dir_path))
            return dict_Result_PathList

        for filename in os.listdir(search_dir_path):
            for re_pattern in re_target_list:
                if re.search(re_pattern, filename):
                    break
            else:
                continue

            remove_path = os.path.join(search_dir_path, filename)
            result = cls.removeFile(remove_path)
            dict_Result_PathList.setdefault(result, []).append(remove_path)

        return dict_Result_PathList

    @classmethod
    def removeMatchFileRecursively(cls, search_dir_path, re_remove_list):
        logger.debug("remove matched file: search-root=%s, re=%s" % (
            search_dir_path, str(re_remove_list)))

        dict_Result_PathList = {}

        if not os.path.isdir(search_dir_path):
            logger.debug("directory not found: " + search_dir_path)
            return dict_Result_PathList

        re_compile_list = [
            re.compile(re_pattern)
            for re_pattern in re_remove_list
            if common.isNotEmptyString(re_pattern)
        ]

        for dir_path, _dir_name_list, filename_list in os.walk(search_dir_path):
            for filename in filename_list:
                for re_pattern in re_compile_list:
                    if re_pattern.search(filename):
                        break
                else:
                    continue

                remove_path = os.path.join(dir_path, filename)
                result = cls.removeFile(remove_path)
                dict_Result_PathList.setdefault(result, []).append(remove_path)

        common.debugDict(dict_Result_PathList, locals())

        return dict_Result_PathList

    @classmethod
    def removeMatchDirectory(cls, search_dir_path, re_target_list):
        import shutil

        logger.debug("search-root=%s, target=%s" % (
            search_dir_path, str(re_target_list)))

        dict_Result_PathList = {}

        for dir_path, dir_name_list, _filename_list in os.walk(search_dir_path):
            for dir_name in dir_name_list:
                for re_text in re_target_list:
                    if re_text is None:
                        logger.debug("null regular expression")
                        continue
                    if re.search(re_text, dir_name):
                        break
                else:
                    continue

                remove_path = os.path.join(dir_path, dir_name)
                result = False
                logger.debug("remove directory: " + remove_path)
                if not cls.__dry_run:
                    try:
                        shutil.rmtree(remove_path, False)
                        result = True
                    except Exception, e:
                        logger.exception(e)

                dict_Result_PathList.setdefault(result, []).append(remove_path)

        return dict_Result_PathList


def validatePath(path):
    common.verifyVarType(common.isString, path)

    work_path = os.path.normpath(path)

    if common.isEmptyString(path):
        raise InvalidFilePathError("null path")

    if all([w == ".." for w in work_path.split(os.path.sep)]):
        raise InvalidFilePathError(work_path)

    if work_path in ("/", "//"):
        raise InvalidFilePathError("root path")


def check_file_existence(path):
    """
    return value:
            FileType

    raise:
            - InvalidFilePathError
            - FileNotFoundError
            - EmptyFileError
            - RuntimeError
    """

    validatePath(path)

    if not os.path.exists(path):
        raise FileNotFoundError(path)

    if os.path.isfile(path):
        logger.debug("file found: " + path)
        if os.path.getsize(path) <= 0:
            raise EmptyFileError(path)

        return FileType.FILE

    if os.path.isdir(path):
        logger.debug("directory found: " + path)
        return FileType.DIRECTORY

    if os.path.islink(path):
        logger.debug("link found: " + path)
        return FileType.LINK

    raise RuntimeError()


def getFileNameFromPath(path):
    """
    フルパスから拡張子を除くファイル名を返す。
    """

    common.verifyVarType(common.isString, path)

    if common.isEmptyString(path):
        return ""

    path = path.strip().strip(os.path.sep)

    return os.path.splitext(os.path.basename(path))[0]


def getRelativePath(target_path_org, root_dir_path):
    import platform

    major, minor, _patchlevel = platform.python_version_tuple()

    if common.isEmptyString(target_path_org):
        logger.debug("empty target path")
        return None

    target_path = target_path_org.strip()

    if os.path.isfile(root_dir_path):
        root_dir_path = os.path.dirname(root_dir_path)

    if int(major) >= 2 and int(minor) >= 6:
        return os.path.relpath(target_path, root_dir_path)
    else:
        try:
            validatePath(target_path)
        except InvalidFilePathError:
            return None

        target_path_list = os.path.abspath(
            target_path).lstrip(os.path.sep).split(os.path.sep)
        start_path_list = os.path.abspath(root_dir_path).lstrip(
            os.path.sep).split(os.path.sep)

        prefix_path_list = []
        common_path_list = []

        is_match = True
        for target, start in zip(target_path_list, start_path_list):
            if is_match and target != start:
                is_match = False

            if is_match:
                common_path_list.append(target)
            else:
                prefix_path_list.append("..")

        if common.isEmptyListOrTuple(common_path_list):
            return target_path_org

        replace_path = os.path.normpath(
            os.path.sep + os.path.sep.join(common_path_list))
        relative_path = target_path.replace(replace_path, ".")
        if common.isEmptyListOrTuple(prefix_path_list):
            return relative_path

        return os.path.join(os.path.join(*prefix_path_list), relative_path)
        # return os.path.join(".".join(prefix_path_list), relative_path)
        # return relative_path


def findFile(search_root_dir_path, re_pattern_text):
    result = findFileAll(
        search_root_dir_path, os.path.isfile, re_pattern_text, find_count=1)

    if common.isEmptyListOrTuple(result):
        return None

    return result[0]


def findFileAll(
        search_root_dir_path, check_func,
        re_pattern_text, find_count=sys.maxint):

    if common.isEmptyString(search_root_dir_path):
        logger.warn("null directory path")
        return []

    if common.isEmptyString(re_pattern_text):
        logger.warn("null regular expression")
        return []

    re_compile = re.compile(re_pattern_text)
    path_list = []

    for dir_path, dir_name_list, filename_list in os.walk(search_root_dir_path):
        for file_name in dir_name_list + filename_list:
            path = os.path.join(dir_path, file_name)
            if not check_func(path):
                continue

            if re_compile.search(file_name) is None:
                continue

            path_list.append(path)

            if len(path_list) >= find_count:
                return path_list

    if len(path_list) > 0:
        logger.debug("file found: count=%d, files=(%s)" % (
            len(path_list), ", ".join(path_list)))
        return path_list

    logger.debug(
        "%s does not contain '%s'" % (search_root_dir_path, re_pattern_text))

    return []


def findDirectory(search_root_dir_path, re_pattern, find_count=-1):
    result = findFileAll(
        search_root_dir_path, os.path.isdir, re_pattern, find_count=1)

    if common.isEmptyListOrTuple(result):
        return None

    return result[0]


def sanitizeFileName(path, replacement_text=""):
    if not common.isNotEmptyString(path):
        return None

    path = path.strip()
    re_replace = re.compile("[%s]" % re.escape('\:/*?"<>|'))

    return re_replace.sub(replacement_text, path)


def adjustFileName(file_name, replacement_text=""):
    fname = sanitizeFileName(file_name, replacement_text)
    if fname is None:
        return None

    fname = common.RE_SPACE.sub("_", fname)
    return re.sub("[%s]" % (re.escape(",().%")), "", fname)


def parsePermission3Char(permission):
    """
    'rwx' 形式のアクセス権限文字列 permission を8進数形式に変換する

    Return value:
            int
    """

    common.verifyVarType(common.isNotEmptyString, permission)

    if len(permission) != 3:
        raise ValueError(permission)

    permission_int = 0
    if permission[0] == "r":
        permission_int += 4
    if permission[1] == "w":
        permission_int += 2
    if permission[2] == "x":
        permission_int += 1

    return permission_int


def parseLsPermissionText(permission_text):
    """
    parse "ls -l" style permission text: e.g. -rw-r--r--
    """

    common.verifyVarType(common.isNotEmptyString, permission_text)

    match = re.search("[-drwx]+", permission_text)
    if match is None:
        raise ValueError(
            "invalid permission character: " + permission_text)

    if len(permission_text) != 10:
        raise ValueError(
            "invalid permission text length: " + permission_text)

    permission_text = permission_text[1:]

    return int(
        "0" + "".join([
            str(parsePermission3Char(permission_text[i:i + 3]))
            for i in range(0, 9, 3)
        ]),
        base=8)
