# encoding: utf-8

'''
@author: vmmhypervisor
'''

import re
import datetime

from golib.logger import logger
import golib.common as common


class RegularExpression:

    class ISO:
        YEAR = "[1-2][\d]{3}"
        DATE = YEAR + "-[0-1][\d]-[0-3][\d]"
        TIME = "[0-2][0-9]:[0-5][0-9]:[0-5][0-9]"
        DATETIME = " ".join([DATE, TIME])
        TIMEZONE = "[+-][\d]{4}"
        DATETIME_WITH_TZ = " ".join([DATETIME, TIMEZONE])


class Format:

    class ISO8601:

        class Basic:
            DATE = "%Y%m%d"
            TIME = "%H%M%S"
            DATETIME = "T".join([DATE, TIME])

        class Extended:
            DATE = "%Y-%m-%d"
            TIME = "%H:%M:%S"
            DATETIME = "T".join([DATE, TIME])

    class ISO:
        DATE = "%Y-%m-%d"
        TIME = "%H:%M:%S"
        DATETIME = " ".join([DATE, TIME])

    ISO_DATETIME_LIST = [
        ISO.DATETIME,
        ISO8601.Basic.DATETIME,
        ISO8601.Extended.DATETIME,
    ]

    DATETIME_LIST = ISO_DATETIME_LIST + [
        " ".join(["%Y/%m/%d", ISO.TIME]),
    ]

    JST_DATE = "%Y/%m/%d"


class DateTimeRange(common.BaseObject):

    @property
    def start_datetime(self):
        return self.__start_datetime

    @property
    def end_datetime(self):
        return self.__end_datetime

    def __init__(
            self, start_datetime, end_datetime,
            time_format=Format.ISO.DATETIME):

        super(DateTimeRange, self).__init__()
        self.__start_datetime = start_datetime
        self.__end_datetime = end_datetime
        self.time_format = time_format

    def equals(self, rhs):
        return all([
            self.start_datetime == rhs.start_datetime,
            self.end_datetime == rhs.end_datetime,
        ])

    def verifyTimeRange(self):
        """
        raise:
                TypeError
                ValueError
        """

        common.verifyVarType(isDatetime, self.start_datetime)
        common.verifyVarType(isDatetime, self.end_datetime)

        if self.start_datetime > self.end_datetime:
            message = "time inversion found: %s > %s" % (
                str(self.start_datetime), str(self.end_datetime))
            raise ValueError(message)

    def verifyTimeFormat(self):
        common.verifyVarType(common.isString, self.time_format)

        if re.search("\%", self.time_format) is None:
            raise ValueError("invalid time format: " + self.time_format)

    def isValidTimeRange(self):
        try:
            self.verifyTimeRange()
        except (TypeError, ValueError):
            return False

        return True

    def isWithin(self, input_datetime):
        self.verifyTimeRange()
        common.verifyVarType(isDatetime, input_datetime)

        return self.start_datetime <= input_datetime <= self.end_datetime

    def getStartTimeText(self):
        if not isDatetime(self.start_datetime):
            return ""

        self.verifyTimeFormat()

        return self.start_datetime.strftime(self.time_format)

    def getEndTimeText(self):
        if not isDatetime(self.end_datetime):
            return ""

        self.verifyTimeFormat()

        return self.end_datetime.strftime(self.time_format)

    def getOptionString(self):
        self.verifyTimeFormat()

        options_list = []

        if isDatetime(self.start_datetime):
            options_list.extend(
                ["-s", "'%s'" % (self.getStartTimeText())])

        if isDatetime(self.end_datetime):
            options_list.extend(
                ["-e", "'%s'" % (self.getEndTimeText())])

        return " ".join(options_list)

    def getTimeDelta(self):
        """
        Return value:
                datetime.timedelta
        """

        return self.end_datetime - self.start_datetime

    def getDeltaSecond(self):
        dt = self.getTimeDelta()

        return (
            dt.days * getTimeUnitSecondsCoefficient("d") +
            float(dt.seconds) + float(dt.microseconds / (1000.0 ** 2)))

    def toString(self, joint=" - ", time_format=None):
        if not self.isValidTimeRange():
            return ""

        old_time_format = None
        if common.isNotEmptyString(time_format):
            old_time_format = self.time_format
            self.time_format = time_format

        text_list = [
            self.getStartTimeText(),
            self.getEndTimeText(),
        ]
        return_value = (
            joint.join(text_list) +
            " (%s)" % (self.end_datetime - self.start_datetime)
        )

        if common.isNotEmptyString(old_time_format):
            self.time_format = old_time_format

        return return_value

    def squeezeTimeRange(self, datetime_range):
        """
        time inversionを発生させない範疇で時間範囲を狭める。
        """

        start = datetime_range.start_datetime
        if isDatetime(start):
            if not isDatetime(self.start_datetime):
                self.__start_datetime = start
            elif isDatetime(self.end_datetime) and self.isWithin(start):
                self.__start_datetime = max(self.start_datetime, start)

        end = datetime_range.end_datetime
        if isDatetime(end):
            if not isDatetime(self.end_datetime):
                self.__end_datetime = end
            elif isDatetime(self.start_datetime) and self.isWithin(end):
                self.__end_datetime = min(self.end_datetime, end)

    def widenTimeRange(self, datetime_range):
        """
        時間範囲を広げる。
        """

        start = datetime_range.start_datetime
        if isDatetime(start):
            if not isDatetime(self.start_datetime):
                self.__start_datetime = start
            else:
                self.__start_datetime = min(self.start_datetime, start)

        end = datetime_range.end_datetime
        if isDatetime(end):
            if not isDatetime(self.end_datetime):
                self.__end_datetime = end
            else:
                self.__end_datetime = max(self.end_datetime, end)

    def discard(self, discard_percent):
        """
        時間範囲の discard_percent / 2 [%] の時間分、開始・終了時刻をずらす。
        """

        self.verifyTimeRange()

        common.verifyVarType(common.isFloat, discard_percent)

        if discard_percent == 0:
            return True

        if discard_percent < 0:
            raise ValueError(
                "discard_percent must be greater than zero: " +
                str(discard_percent))

        discard_time = self.getTimeDelta() // int(100) * \
            int(discard_percent / 2.0)

        self.__start_datetime += discard_time
        self.__end_datetime -= discard_time

        return True


class TimeMeasure(object):
    import logging

    # Message Format ---
    __MF_START = "<start %s> %s"
    __MF_COMPLETE = "<complete> %s: execution-time=%s"

    @property
    def message(self):
        return self.__message

    def __init__(self, message, log_level=logging.INFO):
        import os
        import socket

        self.start()

        self.__message = message
        self.__log_level = log_level

        # output message ---
        hostname = socket.gethostname()
        try:
            ip_address = socket.gethostbyname(hostname)
        except socket.gaierror:
            ip_address = "na"

        header = "%s(%s) %s$ " % (hostname, ip_address, os.getcwd())
        write_message = self.__MF_START % (
            self.__start_datetime.strftime(Format.ISO.DATETIME),
            header + self.message)
        logger.write(write_message, self.__log_level)

    def start(self):
        self.__start_datetime = datetime.datetime.now()

    def stop(self):
        end_datetime = datetime.datetime.now()
        self.__datetimerange = DateTimeRange(
            self.__start_datetime, end_datetime)

        return self.__datetimerange

    def __del__(self):
        datetimerange = self.stop()
        complete_msg = self.__MF_COMPLETE % (
            self.message,
            datetimerange.toString())
        logger.write(complete_msg, self.__log_level)


def isDatetime(value):
    return value is not None and isinstance(value, datetime.datetime)


def isValidTimeFormat(datetime_string, time_format):
    common.verifyVarType(common.isNotEmptyString, time_format)

    if not common.isNotEmptyString(datetime_string):
        return False

    try:
        datetime.datetime.strptime(datetime_string, time_format)
    except ValueError, e:
        logger.debug(e)
        return False

    return True


def toDateTime(datetime_string, time_format, timezone=""):
    common.verifyVarType(common.isNotEmptyString, time_format)

    try:
        datetime_string = str(datetime_string)
    except UnicodeEncodeError:
        return None

    if not common.isNotEmptyString(datetime_string):
        logger.debug(
            "toDateTime - failed to convert: input=%s (%s), format=%s" % (
                datetime_string, type(datetime_string), time_format))
        return None

    try:
        return_datetime = datetime.datetime.strptime(
            datetime_string.strip(), time_format)
    except ValueError, e:
        logger.debug(e)
        return None

    if common.isEmptyString(timezone):
        return return_datetime

    match = re.search(RegularExpression.ISO.TIMEZONE, timezone)
    if match is not None:
        tz = match.group()
        op = tz[0]
        hours = int(tz[1:3])
        minutes = int(tz[3:5])
        td = datetime.timedelta(hours=hours, minutes=minutes)
        if op == "+":
            return_datetime += td
        else:
            return_datetime -= td

    return return_datetime


def toDateTimeEx(datetime_string, time_format_list):
    time_format = findValidTimeFormat(
        datetime_string, time_format_list)

    if time_format is None:
        raise ValueError(
            "there is no matching time format: valid-format=%s, input=%s" % (
                "|".join(time_format_list), datetime_string))

    return toDateTime(datetime_string, time_format)


def findValidTimeFormat(datetime_string, time_format_list):
    common.debugFunc(findValidTimeFormat, locals())

    for time_format in time_format_list:
        if isValidTimeFormat(datetime_string, time_format):
            return time_format

    return None


def getTimeUnitSecondsCoefficient(unit):
    common.verifyVarType(common.isString, unit)

    if not unit.isalpha():
        raise ValueError("invalid unit: " + unit)

    unit = unit.lower()

    if unit == "s":
        return 1
    if unit == "m":
        return 60
    if unit == "h":
        return 60 ** 2
    if unit == "d":
        return 60 ** 2 * 24
    if unit == "w":
        return 60 ** 2 * 24 * 7

    raise ValueError("unknown unit: " + unit)


def getTimeDeltaSecond(start_datetime, end_datetime):
    common.verifyVarType(isDatetime, start_datetime)
    common.verifyVarType(isDatetime, end_datetime)

    deltatime = (end_datetime - start_datetime)
    return (
        deltatime.days * getTimeUnitSecondsCoefficient("d") +
        float(deltatime.seconds) +
        float(deltatime.microseconds / (1000.0 ** 2)))


def convertHumanReadableToSecond(readable_time):
    common.verifyVarType(common.isString, readable_time)

    if common.isEmptyString(readable_time):
        raise ValueError("empty input")

    size = readable_time[:-1]
    unit = readable_time[-1]
    if not common.isFloat(size):
        raise ValueError("invalid value: '%s'" % (size))
    size = float(size)

    if size < 0:
        raise ValueError("minus size")

    return size * getTimeUnitSecondsCoefficient(unit)
