# encoding: utf-8

'''
@author: vmmhypervisor
'''

import re
import datetime

import six

import golib.common as common
import golib.sqlite as sql
import golib.gtime as gtime
import golib.gfile as gfile
import golib.statistics
from golib.logger import logger


DISK_SECTOR_SIZE = 512        # [byte]
BLOCK_IO_SIZE = 512        # [byte]
SUMARRY_STATISTICS_TYPE = golib.statistics.TRIMED_MEAN

RAW_SAR_FILE_NAME_PREFIX = "raw_sar"
EXT_SAR_BIN_DATA = ".dat"

RAW_PIDSTAT_FILE_NAME_PREFIX = "pidstat"


# Table Name Suffix
TNP_IrqPerTps = "IRQ per TPS"

# Field Name (pidstat) ---
FN_COMMAND = "Command"
FN_PID = "PID"

# Field Name (sar) ---
FN_KeyPriority = "Key Priority"
FN_GroupId = "group_id"
FN_GroupPriority = "Group Priority"

FN_SysstatFieldOrder = "sysstat_order"
FN_Table_Name = "table_name"

FN_SarCpuUtil = "%user+%system"  # del
FN_ReadKbPerSec = "read_kB/s"
FN_WriteKbPerSec = "write_kB/s"
FN_TotalKbPerSec = "total_kB/s"

FN_ReadSizePerTransfer = "read_kB/trans"
FN_WriteSizePerTransfer = "write_kB/trans"
FN_IoSizePerTransfer = "total_kB/trans"
FN_TotalAwait = "await_x_tps"
FN_TotalServiceTime = "svctm_x_tps"

#FN_IoSizePerTransfer    = "io_kb/trans"
FN_PercentageOfRead = "%read"
FN_AmountRate = "%amount"
FN_TotalPacketPerSec = "total_pck/s"

FN_WriteKbPerPacket = "write_kB/pck"
FN_ReadKbPerPacket = "read_kB/pck"
FN_TotalKbPerPacket = "total_kB/pck"

# /proc/meminfo ---
TN_MeminfoKb = "meminfo kB"
TN_MeminfoHugePages = "meminfo HugePages"

FN_MeminfoKb = TN_MeminfoKb
FN_MeminfoHugePages = TN_MeminfoHugePages

# Database Name ---
DN_RawSarData = "Raw Sar Data"
DN_SarStatistics = "Sar Statistics"
DN_RawPidstatData = "Raw Pidstat Data"

PIDSTAT_DB_NAME = "pidstat"

# Database Version ---
DV_RawSarData = "1.4.1"
DV_SarStatistics = "1.6.1"
DV_RawPidstatData = "1.0.0"

PIDSTAT_DB_VERSION = "0.1.1"

# ---
TABLE_NAME_SEP = " "

# Table Name Suffix
TNP_PerCpu = " per CPU"
TNP_PerType = " per Type"
TNP_Timeline = " Timeline"
TNP_OrderBy = " Order By"
TNP_Summary = " Summary"
TNP_Info = " Information"

# Table Name
TN_CpuUtil = "CPU Utilization"
TN_Memory = "Memory"
TN_TaskCreated = "Task"
TN_DiskIO = "Disk I/O"
TN_Interrupt = "Interrupts"
TN_InterruptEx = "Interrupts Ex"
TN_BlockIO = "Block I/O"
TN_MemoryUtil = "Memory and Swap Utilization"
TN_Queue = "Queue"
TN_Paging = "Paging"
TN_inode = "inode"
TN_Swap = "Swap"
TN_Comparison = "Summary"
TN_SystemSwitch = "System Switching"

TN_NetworkPrefix = "Network I/O"
TN_NetworkFormat = TN_NetworkPrefix + " %s"
TN_NetworkDev = TN_NetworkFormat % ("DEV")
TN_NetworkEdev = TN_NetworkFormat % ("EDEV")

#TN_TtyDevice        = "TTY Device"

TN_CpuInfo = "CPU" + TNP_Info
TN_InterruptInfo = TN_Interrupt + TNP_Info
TN_DiskInformation = TN_DiskIO + TNP_Info

TN_EnvironmentInfo = "Converter Environment Information"

# Table Name ---
TN_InterruptPerType = TN_Interrupt + TNP_PerType
TN_CpuInterruptSummary = "%s - %s%s" % (
    TN_CpuUtil, TN_Interrupt, TNP_Summary)
TN_Time = "Time"

# Field Name
FN_Cpu = "CPU"

FN_DiskInformationId = "disk info id"
FN_MajorDeviceId = "major device id"
FN_MinorDeviceId = "minor device id"
FN_DiskDevice = "DEV"
FN_DiskStats = "device"
FN_DiskLdev = "LDEV"
FN_DiskHdlm = "HDevName"
FN_DiskLun = "LUN"
FN_FileSystem = "file_system"

FN_IrqType = "irq_type"
FN_IrqNumber = "irq_no"
FN_IrqSarName = "sar_name"
FN_IrqDeviceName = "device_name"

FN_CPU_IDLE = "%idle"
FN_UserRate = "%user_rate"

IRQ_TYPE_Network = "network"
IRQ_TYPE_Storage = "disk"
IRQ_TYPE_Timer = "timer"
IRQ_TYPE_Virtualize = "virt"
IRQ_TYPE_Etc = "etc"
dict_IrqType_Priority = {
    IRQ_TYPE_Storage: 0,
    IRQ_TYPE_Network: 1,
    IRQ_TYPE_Virtualize: 2,
    IRQ_TYPE_Etc: 3,
}

NO_PRIMARY_KEY = "NO_PRIMARY_KEY"
DEFAULT_PRIMARY_KEY_FIELD = NO_PRIMARY_KEY

# Sysstat Field List
SFL_CPU = frozenset(["%user", "%system"])


# class ---

class SysstatVersion:
    VER_8 = "080102"  # 8.1.2
    VER_11 = "110001"  # 11.0.1
    UNKNOWN = "unknown"

    LATEST_SUPPORTED = VER_11

    VERSION_LIST = [VER_11, VER_8]


class SysstatInfo:
    TABLE_NAME = "Sysstat Information"

    class Field:
        SYSSTAT_HEADER = "Sysstat Header"
        HOSTNAME = "Hostname"
        SYSSTAT_VERSION = "Sysstat Version"

    @classmethod
    def getHostname(cls, con):
        if not con.hasTable(cls.TABLE_NAME):
            return "unknown"

        return con.getValue(
            select=sql.SqlQuery.to_attr_str(common.AN_GeneralValue),
            table=cls.TABLE_NAME,
            where=sql.SqlQuery.make_where(
                common.AN_GeneralKey, cls.Field.HOSTNAME))

    @classmethod
    def getSysstatVersion(cls, con):
        if not con.hasTable(cls.TABLE_NAME):
            return SysstatVersion.UNKNOWN

        return con.getValue(
            select=sql.SqlQuery.to_attr_str(common.AN_GeneralValue),
            table=cls.TABLE_NAME,
            where=sql.SqlQuery.make_where(
                common.AN_GeneralKey, cls.Field.SYSSTAT_VERSION))

    @classmethod
    def makeSysstatInfoTable(
            cls, con_out, sar_header_line, sysstat_version=SysstatVersion.UNKNOWN):

        sampling_hostname, _sampling_start_date    = \
            SysstatParser.parseSysstatHeader(sar_header_line)

        if con_out.hasTable(cls.TABLE_NAME):
            return True

        value_matrix = [
            [cls.Field.SYSSTAT_HEADER, sar_header_line],
            [cls.Field.HOSTNAME, sampling_hostname],
            [cls.Field.SYSSTAT_VERSION, sysstat_version],
        ]
        args = {
            "table_name": cls.TABLE_NAME,
            "attribute_name_list": common.KEY_VALUE_HEADER,
            "data_matrix": value_matrix,
        }

        return con_out.createTableWithDataEx(**args)


class SysstatObjectBase(object):

    @property
    def field_list(self):
        return []

    @property
    def modify_field_list(self):
        return self.field_list


class TaskCreated:
    TABLE_NAME = TN_TaskCreated

    @classmethod
    def factory(cls, sysstat_version):
        class sysstat_8(SysstatObjectBase):

            @property
            def field_list(self):
                return ['proc/s']

        class sysstat_11(SysstatObjectBase):
            pass

        if sysstat_version == SysstatVersion.VER_11:

            return sysstat_11()

        return sysstat_8()


class SystemSwitch:
    TABLE_NAME = TN_SystemSwitch

    @classmethod
    def factory(cls, sysstat_version):
        class sysstat_8(SysstatObjectBase):

            @property
            def field_list(self):
                return ['cswch/s']

        class sysstat_11(SysstatObjectBase):
            pass

        if sysstat_version == SysstatVersion.VER_11:

            return sysstat_11()

        return sysstat_8()


class TaskCreatedAndSystemSwitch:
    TABLE_NAME = TaskCreated.TABLE_NAME + SystemSwitch.TABLE_NAME

    @classmethod
    def factory(cls, sysstat_version):
        class sysstat_8(SysstatObjectBase):
            pass

        class sysstat_11(SysstatObjectBase):

            @property
            def field_list(self):
                return ['proc/s', 'cswch/s']

        if sysstat_version == SysstatVersion.VER_11:
            return sysstat_11()

        return sysstat_8()


class Swap:
    TABLE_NAME = "Swap"

    @classmethod
    def factory(cls, sysstat_version):
        class sysstat_8(SysstatObjectBase):

            @property
            def field_list(self):
                return ['pswpin/s', 'pswpout/s']

        return sysstat_8()


class Paging:
    TABLE_NAME = TN_Paging

    @classmethod
    def factory(cls, sysstat_version):
        class sysstat_8(SysstatObjectBase):

            @property
            def field_list(self):
                return [
                    'pgpgin/s', 'pgpgout/s', 'fault/s', 'majflt/s',
                    'pgfree/s', 'pgscank/s', 'pgscand/s', 'pgsteal/s', '%vmeff',
                ]

        return sysstat_8()


class BlockIo:
    TABLE_NAME = TN_BlockIO

    @classmethod
    def factory(cls, sysstat_version):
        class sysstat_8(SysstatObjectBase):
            TPS = 'tps'
            READ_TPS = 'rtps'
            WRITE_TPS = 'wtps'
            BLOCK_READ_PER_SEC = 'bread/s'
            BLOCK_WRITE_PER_SEC = 'bwrtn/s'

            @property
            def field_list(self):
                return [
                    self.TPS,
                    self.READ_TPS,
                    self.WRITE_TPS,
                    self.BLOCK_READ_PER_SEC,
                    self.BLOCK_WRITE_PER_SEC,
                ]

        return sysstat_8()


class MemoryUtil:
    TABLE_NAME = TN_MemoryUtil

    @classmethod
    def factory(cls, sysstat_version):
        if sysstat_version == SysstatVersion.VER_11:
            class sysstat_11(SysstatObjectBase):

                @property
                def field_list(self):
                    return [
                        'kbmemfree', 'kbmemused', '%memused', 'kbbuffers',
                        'kbcached', 'kbcommit', '%commit', 'kbactive',
                        'kbinact', 'kbdirty',
                    ]

            return sysstat_11()

        class sysstat_8(SysstatObjectBase):

            @property
            def field_list(self):
                return [
                    'kbmemfree', 'kbmemused', '%memused', 'kbbuffers',
                    'kbcached', 'kbswpfree', 'kbswpused', '%swpused', 'kbswpcad'
                ]

        return sysstat_8()


class Inode:
    TABLE_NAME = TN_inode

    @classmethod
    def factory(cls, sysstat_version):
        class sysstat_8(SysstatObjectBase):

            @property
            def field_list(self):
                return ['dentunusd', 'file-nr', 'inode-nr', 'pty-nr']

        return sysstat_8()


class Queue:
    TABLE_NAME = TN_Queue

    @classmethod
    def factory(cls, sysstat_version=SysstatVersion.LATEST_SUPPORTED):
        if sysstat_version == SysstatVersion.VER_11:
            class sysstat_11(SysstatObjectBase):

                @property
                def field_list(self):
                    return [
                        'runq-sz', 'plist-sz', 'ldavg-1', 'ldavg-5', 'ldavg-15', 'blocked'
                    ]

            return sysstat_11()

        class sysstat_8(SysstatObjectBase):

            @property
            def field_list(self):
                return [
                    "runq-sz", "plist-sz", "ldavg-1", "ldavg-5", "ldavg-15",
                ]

        return sysstat_8()


class DiskIo:
    TABLE_NAME = TN_DiskIO

    @classmethod
    def factory(cls, sysstat_version):
        class sysstat_8(SysstatObjectBase):
            DEVICE = 'DEV'
            TPS = 'tps'
            READ_SECTOR_PER_SEC = 'rd_sec/s'
            WRITE_SECTOR_PER_SEC = 'wr_sec/s'
            AWAIT = 'await'
            SERVICE_TIME = 'svctm'

            @property
            def field_list(self):
                return [
                    'DEV', self.TPS,
                    self.READ_SECTOR_PER_SEC,
                    self.WRITE_SECTOR_PER_SEC,
                    'avgrq-sz', 'avgqu-sz',
                    self.AWAIT, self.SERVICE_TIME, '%util'
                ]

        return sysstat_8()


class NetworkIoDev:
    TABLE_NAME = TN_NetworkDev

    @classmethod
    def factory(cls, sysstat_version=SysstatVersion.LATEST_SUPPORTED):
        class sysstat_8(SysstatObjectBase):
            RECV_PACKET_PER_SEC = 'rxpck/s'
            SEND_PACKET_PER_SEC = 'txpck/s'
            RECV_KB_PER_SEC = 'rxkB/s'
            SEND_KB_PER_SEC = 'txkB/s'

            @property
            def field_list(self):
                return [
                    'IFACE',
                    self.RECV_PACKET_PER_SEC, self.SEND_PACKET_PER_SEC,
                    self.RECV_KB_PER_SEC, self.SEND_KB_PER_SEC,
                    'rxcmp/s', 'txcmp/s', 'rxmcst/s'
                ]

        if sysstat_version == SysstatVersion.VER_11:
            class sysstat_11(sysstat_8):

                @property
                def field_list(self):
                    return super(sysstat_11, self).field_list + ['%ifutil']

            return sysstat_11()

        return sysstat_8()


class CpuUtilization:
    TABLE_NAME = "CPU Utilization"

    @classmethod
    def factory(cls, sysstat_version=SysstatVersion.LATEST_SUPPORTED):
        class sysstat_8(SysstatObjectBase):
            CPU = "CPU"
            USER_UTIL = "%user"
            SYSTEM_UTIL = "%system"
            IO_WAIT = '%iowait'
            CPU_UTIL = USER_UTIL + "+" + SYSTEM_UTIL

            @property
            def field_list(self):
                return [
                    self.CPU,
                    self.USER_UTIL,
                    '%nice',
                    self.SYSTEM_UTIL,
                    self.IO_WAIT,
                    '%steal',
                    '%idle',
                ]

            @property
            def modify_field_list(self):
                return self.field_list + [self.CPU_UTIL]

        class sysstat_11(sysstat_8):
            USER_UTIL = "%usr"
            SYSTEM_UTIL = "%sys"
            CPU_UTIL = USER_UTIL + "+" + SYSTEM_UTIL

            @property
            def field_list(self):
                return [
                    self.CPU,
                    self.USER_UTIL,
                    '%nice',
                    self.SYSTEM_UTIL,
                    self.IO_WAIT,
                    '%steal',
                    "%irq",
                    "%soft",
                    "%guest",
                    "%gnice",
                    '%idle',
                ]

            @property
            def modify_field_list(self):
                return self.field_list + [self.CPU_UTIL]

        if sysstat_version == SysstatVersion.VER_11:
            return sysstat_11()

        return sysstat_8()


# ---
"""
class Paging:
    class Field:
        SET    = frozenset([
            'pgpgin/s', 'pgpgout/s', 'fault/s', 'majflt/s',
            'pgfree/s', 'pgscank/s', 'pgscand/s', 'pgsteal/s', '%vmeff',
        ])

    @classmethod
    def getTableName(cls):
        return TN_Paging


class BlockIo:
    class Field:
        TPS            = 'tps'
        READ_TPS    = 'rtps'
        WRITE_TPS    = 'wtps'
        BLOCK_READ_PER_SEC    = 'bread/s'
        BLOCK_WRITE_PER_SEC    = 'bwrtn/s'
        
        SET    = frozenset([
            TPS, READ_TPS, WRITE_TPS, BLOCK_READ_PER_SEC, BLOCK_WRITE_PER_SEC])

    @classmethod
    def getTableName(cls):
        return TN_BlockIO
        
        
class MemoryUtil:
    class Field:
        SET    = frozenset([
            'kbmemfree', 'kbmemused', '%memused', 'kbbuffers',
            'kbcached', 'kbswpfree', 'kbswpused', '%swpused', 'kbswpcad'])
        
    @classmethod
    def getTableName(cls):
        return TN_MemoryUtil


class DiskIo:
    class Field:
        DEVICE                    = 'DEV'
        TPS                        = 'tps'
        READ_SECTOR_PER_SEC        = 'rd_sec/s'
        WRITE_SECTOR_PER_SEC    = 'wr_sec/s'
        AWAIT            = 'await'
        SERVICE_TIME    = 'svctm'
        
        SET    = frozenset([
            'DEV', TPS, READ_SECTOR_PER_SEC, WRITE_SECTOR_PER_SEC,
            'avgrq-sz', 'avgqu-sz', AWAIT, SERVICE_TIME, '%util'])

    @classmethod
    def getTableName(cls):
        return TN_DiskIO


class NetworkIoDev:
    class Field:
        RECV_PACKET_PER_SEC    = 'rxpck/s'
        SEND_PACKET_PER_SEC    = 'txpck/s'
        RECV_KB_PER_SEC        = 'rxkB/s'
        SEND_KB_PER_SEC        = 'txkB/s'
        
        SET    = frozenset([
            'IFACE', RECV_PACKET_PER_SEC, SEND_PACKET_PER_SEC,
            RECV_KB_PER_SEC, SEND_KB_PER_SEC, 'rxcmp/s', 'txcmp/s', 'rxmcst/s'])

    @classmethod
    def getTableName(cls):
        return TN_NetworkDev

"""


class CpuUtil:

    class Field:
        CPU = "CPU"
        USER_UTIL = "%user"
        SYSTEM_UTIL = "%system"
        IO_WAIT = '%iowait'
        CPU_UTIL = USER_UTIL + "+" + SYSTEM_UTIL

        ORG = frozenset([
            CPU, USER_UTIL, SYSTEM_UTIL, '%nice', IO_WAIT, '%steal', '%idle'])
        SET = frozenset([
            CPU, USER_UTIL, SYSTEM_UTIL, CPU_UTIL, '%nice', IO_WAIT, '%steal', '%idle'])

    @classmethod
    def getTableName(cls):
        return TN_CpuUtil

    @classmethod
    def getPrimaryKeyFieldList(cls):
        return [FN_Cpu]


def getTableNameFromField(field, sysstat_version):
    sysstat_class_list = [
        CpuUtilization, BlockIo, DiskIo, NetworkIoDev, MemoryUtil, Queue,
    ]

    for sysstat_class in sysstat_class_list:
        sysstat_obj = sysstat_class.factory(sysstat_version)
        if field in sysstat_obj.modify_field_list:
            return sysstat_class.TABLE_NAME

    raise ValueError(field)

    return None


class Unit():
    BYTE = "byte"
    KB = "kB"

    BYTE_PER_SEC = "byte/sec"
    KB_PER_SEC = "kB/s"

    KB_PER_TRANS = "kB/transport"
    KB_PER_PACKET = "kB/packet"

    PACKET_PER_SEC = "packet/s"
    BLOCK_PER_SEC = "block/s"
    PAGE_PER_SEC = "page/s"

    PERCENT = "percent"
    OPS = "ops"
    MILLISECOND = "msec"

    # sar ---
    TRANSFER_PER_SEC = "tps"

    # pfm ---
    IOPS = "iops"
    MB = "mB"
    MB_PER_SEC = "mB/s"
    COUNT = "count"
    MICRO_SEC = "micro_sec"

    DIMENSIONLESS = "dimensionless"


class SysstatParser:

    class SysstatGroupData(common.BaseObject):

        @property
        def group_index(self):
            return self.__group_index

        @property
        def group_name(self):
            return self.__group_name

        @property
        def label(self):
            return self.__label

        @property
        def re_pattern(self):
            return self.__re_pattern

        def __init__(self, index, group_name, label, re_pattern):
            self.__group_index = index
            self.__group_name = group_name
            self.__label = label
            self.__re_pattern = re.compile(re_pattern)

        def __repr__(self):
            # return "index=%d, group-name=%s, re=%s" % (
            #    self.group_index, self.group_name, self.re_pattern)
            return self.__class__.__name__ + common.dumpDict(self.__dict__)

    class Label:
        TRANS_PER_SEC = "transfers/sec"
        SIZE_PER_TRANS = "size/transfer"
        SIZE_PER_PACKET = "size/packet"

    __SYSSTAT_GROUP_DATA_LIST = [
        # sar
        SysstatGroupData(
            1, Unit.KB_PER_SEC, Label.TRANS_PER_SEC, "kB.*/s"),
        SysstatGroupData(
            2, Unit.TRANSFER_PER_SEC, Label.TRANS_PER_SEC, "tps$"),
        SysstatGroupData(3, Unit.PERCENT, Unit.PERCENT, "%"),
        SysstatGroupData(
            4, Unit.KB_PER_TRANS, Label.SIZE_PER_TRANS, "kB/trans$"),
        SysstatGroupData(
            5, Unit.KB_PER_PACKET, Label.SIZE_PER_PACKET, "kB/pck$"),
        SysstatGroupData(
            6, Unit.BLOCK_PER_SEC, Unit.BLOCK_PER_SEC, "^b[rw].*/s$"),
        SysstatGroupData(
            7, Unit.PACKET_PER_SEC, Unit.PACKET_PER_SEC, "pck/s$"),
        SysstatGroupData(
            8, Unit.PAGE_PER_SEC, Unit.PAGE_PER_SEC, "^pg.*/s|pg/s$"),
        SysstatGroupData(9, "load_average", "load_average", "^ldavg"),
        SysstatGroupData(10, "size", "size", "-sz$"),
        SysstatGroupData(11, "sector/s", "sector/s", "sec/s"),
        SysstatGroupData(12, Unit.KB, Unit.KB, "^kb"),
        SysstatGroupData(
            13, Unit.MILLISECOND, Unit.MILLISECOND, "^await|^svctm"),
        SysstatGroupData(
            14, Unit.DIMENSIONLESS, Unit.DIMENSIONLESS, "^value$"),

        # pidstat
        SysstatGroupData(20, Unit.KB, Unit.KB, "^RSS|^VSZ"),
        SysstatGroupData(21, "fault", "fault/s", "^fault/s$|flt/s$"),
        SysstatGroupData(22, "context_switch", "switch/s", "swch/s$"),

        # meminfo
        SysstatGroupData(30, Unit.KB, Unit.KB, re.escape(FN_MeminfoKb)),
        SysstatGroupData(
            31, "pages", "pages", re.escape(FN_MeminfoHugePages)),

        # etc
        SysstatGroupData(99, "etc", "etc", "."),
    ]

    @classmethod
    def getSysstatGroupByGroupName(cls, group_name):
        if common.isEmptyString(group_name):
            return None

        for group_data in cls.__SYSSTAT_GROUP_DATA_LIST:
            if group_data.group_name == group_name:
                return group_data

        return None

    @classmethod
    def getSysstatGroupByFieldName(cls, field):
        if common.isEmptyString(field):
            return None

        for group_data in cls.__SYSSTAT_GROUP_DATA_LIST:
            if group_data.re_pattern.search(field):
                return group_data

        return None

    @classmethod
    def parseSysstatHeader(cls, header_line):
        """
        header_line: LC_ALL=C output

        raise:
                - ValueError
        """

        common.verifyVarType(common.isString, header_line)

        sar_header = header_line.strip()
        item_list = common.RE_SPACE.split(sar_header)
        if len(item_list) < 3:
            raise ValueError("invalid header (too few items): " + header_line)

        hostname = re.sub("[%s]" % (re.escape("()")), "", item_list[2])
        #hostname    = common.getListItem(split_list, 2).lstrip("(").rstrip(")")
        date_text = item_list[3]

        # LC_ALL=C ---
        date_format = "%m/%d/%y"
        try:
            systat_date = gtime.toDateTime(date_text, date_format).date()
        except AttributeError:
            pass
        else:
            return (hostname, systat_date)

        # en_US.UTF-8 ---
        date_format = "%m/%d/%Y"
        try:
            systat_date = gtime.toDateTime(date_text, date_format).date()
        except AttributeError:
            pass
        else:
            return (hostname, systat_date)

        raise ValueError(header_line)

    @classmethod
    def splitSysstatGroup(cls, field_list):
        dict_Group_FieldList = {}

        for field in field_list:
            group_data = cls.getSysstatGroupByFieldName(field)
            if group_data is None:
                continue

            dict_Group_FieldList.setdefault(
                group_data.group_name, []).append(field)

        common.debugDict(dict_Group_FieldList, locals())

        return dict_Group_FieldList


class Formatter(common.BaseObject):

    def __init__(
            self, start_date, start_time, max_datetime=datetime.datetime.max):
        self.__last_date = start_date
        self.__last_time = start_time

        if gtime.isDatetime(max_datetime):
            self.__max_datetime = max_datetime
        else:
            self.__max_datetime = datetime.datetime.max

    def formattingSysstatLineList(self, sysstat_data_line_list, min_field_len):
        SYSSTAT_TIME_FORMAT = "%H:%M:%S"

        value_matrix = []
        min_datetime = None
        max_datetime = None

        sampling_date = self.__last_date
        prev_time = self.__last_time

        dict_TimeText_DateTimeCache = {}
        dict_TimeText_TimeCache = {}

        for line_idx, data_line in enumerate(sysstat_data_line_list):
            line = data_line.strip()
            line_item_list = common.RE_SPACE.split(line)
            if len(line_item_list) < min_field_len:
                logger.warn(
                    "skip invalid line (%d): '%s'" % (line_idx, line))
                continue

            time_text = line_item_list[0]
            sampling_time = dict_TimeText_TimeCache.get(time_text)
            if sampling_time is None:
                # update time cache
                sampling_time = datetime.datetime.strptime(
                    time_text, SYSSTAT_TIME_FORMAT).time()
                dict_TimeText_TimeCache[time_text] = sampling_time

                # 日またぎ処理
                if prev_time is not None:
                    if sampling_time < prev_time:
                        sampling_date = sampling_date + \
                            datetime.timedelta(days=1)
                        dict_TimeText_DateTimeCache = {}        # flush cache

            prev_time = sampling_time

            sampling_datetime = dict_TimeText_DateTimeCache.get(time_text)
            if sampling_datetime is None:
                # update datetime cache
                sampling_datetime = datetime.datetime.combine(
                    sampling_date, sampling_time)
                dict_TimeText_DateTimeCache[time_text] = sampling_datetime

                # update time range
                if min_datetime is None:
                    min_datetime = sampling_datetime
                else:
                    min_datetime = min(min_datetime, sampling_datetime)

                if max_datetime is None:
                    max_datetime = sampling_datetime
                else:
                    max_datetime = max(max_datetime, sampling_datetime)

            if sampling_datetime > self.__max_datetime:
                break

            #line_item_list[0]    = sampling_datetime
            line_item_list[0] = sampling_datetime.strftime(
                sql.DB_DATETIME_FORMAT)
            value_matrix.append(line_item_list)

        self.__last_date = max_datetime.date()
        self.__last_time = max_datetime.time()

        return (gtime.DateTimeRange(min_datetime, max_datetime), value_matrix)


class SysstatParseOption(common.BaseObject):

    def __init__(self):
        super(SysstatParseOption, self).__init__()

        self.table_name = ""
        self.primary_key_field_list = [NO_PRIMARY_KEY]
        self.grouping = True
        self.discard_rate = 0        # [%]

        # show
        self.sort_key_list = None
        self.is_reverse_sort = None
        self.show_field_list = None
        self.show_style = "omit"

    def allValidPrimaryKeyField(self):
        return allPrimaryKey(self.primary_key_field_list)

    def getPrimaryKeyText(self):
        return ", ".join([
            str(key_field) for key_field in self.primary_key_field_list])

    def getPrimaryKeyAttrList(self, con):
        primary_key_field_list = []

        if self.allValidPrimaryKeyField():
            for primary_key_field in self.primary_key_field_list:
                if con.hasAttribute(self.table_name, primary_key_field):
                    primary_key_field_list.append(primary_key_field)
                else:
                    logger.warn("field not found: " + primary_key_field)

        return primary_key_field_list


class Sar:

    @classmethod
    def initialize(cls, sys_wrapper):
        cls.__sys_wrapper = sys_wrapper

    @classmethod
    def verifyBinDataFile(cls, file_path):
        gfile.check_file_existence(file_path)

        file_type = gfile.FileTypeChecker.get_file_type(file_path)

        if file_type != gfile.FileTypeChecker.FileType.BINARY:
            raise IOError("invalid sar bin file: file-type=%s" % (file_type))

        common.checkInstallCommand(["sar"])

        command = "sar -f %s | head" % (file_path)
        sar = cls.__sys_wrapper.popenCommand(command)
        ret_stdout, _ret_stderr = sar.communicate()
        return_code = sar.returncode
        if return_code != 0:
            raise RuntimeError("%s = %d" % (command, return_code))

        buf = ret_stdout.splitlines()
        if len(buf) < 4:
            raise IOError("invalid sar binary file: " + file_path)

        SysstatParser.parseSysstatHeader(buf[0])

        return return_code

    @classmethod
    def isValidBinDataFile(cls, file_path):
        return_value = True

        try:
            return_value = cls.verifyBinDataFile(file_path)
        except gfile.RECOVERABLE_ERROR_LIST:
            raise
        except (IOError, ValueError):
            # not a sar binary file
            return False

        return return_value == 0

    @classmethod
    def verifyTextDataFile(cls, file_path):
        gfile.check_file_existence(file_path)

        if not gfile.FileTypeChecker.is_text_file(file_path):
            raise IOError("invalid sar text file: file-type=%s" % (
                gfile.FileTypeChecker.get_file_type(file_path)))

        buf = open(file_path).readlines()
        if len(buf) < 4:
            raise IOError("invalid sar text file. too few lines: " + file_path)

        SysstatParser.parseSysstatHeader(buf[0])

        return True

    @classmethod
    def isValidTextDataFile(cls, file_path):
        return_value = True

        try:
            return_value = cls.verifyTextDataFile(file_path)
        except gfile.RECOVERABLE_ERROR_LIST:
            raise
        except (IOError, ValueError), e:
            # not a sar text file
            logger.warn(file_path + ": " + str(e))
            return False

        return return_value

    @classmethod
    def startSampling(
            cls, sar_path, sampling_interval_sec, sampling_time_sec=-1):
        if not common.isInteger(sampling_interval_sec):
            logger.error(
                "not integer sampling interval: " + str(sampling_interval_sec))
            return None

        if not common.isInteger(sampling_time_sec):
            logger.error(
                "not integer sampling time: " + str(sampling_time_sec))
            return None

        sampling_count = common.safeDivision(
            sampling_time_sec, sampling_interval_sec)
        if common.isNaN(sampling_count):
            logger.error(
                "invalid arguments: %s, %s" % (
                    str(sampling_interval_sec), str(sampling_time_sec)))
            return None

        command = "sar -A -I XALL -o %s %d %d > /dev/null 2>&1" % (
            sar_path, sampling_interval_sec, sampling_count)

        return cls.__sys_wrapper.popenCommand(command)


def isPrimaryKey(key):
    return all([
        key is not NO_PRIMARY_KEY,
        common.isNotEmptyString(key)
    ])


def allPrimaryKey(key_list):
    if common.isEmptyList(key_list):
        return False

    return all([
        isPrimaryKey(key) for key in key_list])


def getOutputSysstatTableNameList(options, irq_table_name=TN_Interrupt):
    import golib.option

    plot_table_name_list = []

    if common.isNotEmptyString(options.config):
        dict_config = golib.loader.JsonLoader.load(options.config)
        if dict_config:
            return six.iterkeys(dict_config)

    dict_TableName_SarOption = golib.option.getDict_TableName_SarOption(
        options)

    for table_name, values in six.iteritems(dict_TableName_SarOption):
        _, is_output = values

        if not is_output:
            continue

        if table_name == TN_Interrupt:
            table_name = irq_table_name
        elif table_name == TN_NetworkPrefix:
            for network_arg in options.network_arg_list:
                network_table_name = TN_NetworkFormat % (network_arg)
                plot_table_name_list.append(network_table_name)
            continue

        plot_table_name_list.append(table_name)

    logger.debug("table list: " + str(plot_table_name_list))

    return plot_table_name_list


def makeInterruptNameConverterDict(con_raw_sar):
    if con_raw_sar is None:
        logger.error("null database")
        return {}

    if not con_raw_sar.hasTable(TN_InterruptInfo):
        logger.warn("%s table not found" % (TN_InterruptInfo))
        return {}

    dict_SarName_IrqName = {}
    sar_name_list = con_raw_sar.getValueList(
        select="DISTINCT " + sql.SqlQuery.to_attr_str(FN_IrqSarName),
        table=TN_InterruptInfo)
    for sar_name in sar_name_list:
        irq_name = con_raw_sar.getValue(
            select="DISTINCT " + sql.SqlQuery.to_attr_str(FN_IrqDeviceName),
            table=TN_InterruptInfo,
            where=sql.SqlQuery.make_where(FN_IrqSarName, sar_name))

        dict_SarName_IrqName[sar_name] = irq_name

    return dict_SarName_IrqName


def isDiskTableName(table_name):
    return re.match(TN_DiskIO, table_name)


def getDiskTableName(table_name_prefix, field):
    # if field in [FN_DiskDevice, FN_DiskStats]:
    if field in [FN_DiskDevice]:
        return table_name_prefix
    return "%s %s" % (table_name_prefix, field)


def isNetworkInterfaceTableName(table_name):
    return (
        re.search(TN_NetworkPrefix, table_name) and
        re.search("DEV|EDEV", table_name))


class PrimaryKey:
    __dict_TableName_PrimaryKeyField = {
        CpuUtilization.TABLE_NAME: CpuUtil.getPrimaryKeyFieldList(),
        TN_Interrupt: [FN_Cpu],
        TN_InterruptEx: [FN_Cpu, "irq_name", "irq_no"],
        # TN_InterruptEx    : [FN_Cpu],
        TN_NetworkDev: ["IFACE"],
        TN_NetworkEdev: ["IFACE"],
        getDiskTableName(TN_DiskIO, FN_DiskDevice): [FN_DiskStats],
        getDiskTableName(TN_DiskIO, FN_DiskStats): [FN_DiskStats],
        getDiskTableName(TN_DiskIO, FN_DiskLdev): [FN_DiskLdev],
        getDiskTableName(TN_DiskIO, FN_DiskHdlm): [FN_DiskHdlm],
        getDiskTableName(TN_DiskIO, FN_DiskLun): [FN_DiskLun],
        getDiskTableName(TN_DiskIO, FN_FileSystem): [FN_FileSystem],
    }

    @classmethod
    def hasKey(cls, table_name):
        return cls.__dict_TableName_PrimaryKeyField.get(table_name) is not None

    @classmethod
    def getKeyFieldList(
            cls, table_name,
            default_primary_key_list=[DEFAULT_PRIMARY_KEY_FIELD]):

        return cls.__dict_TableName_PrimaryKeyField.get(
            table_name, default_primary_key_list)


def getDateTimeRangeFromTableDev(
        con, table_name, where=None, discard_rate=None):
    """
    raise:
            TypeError
            TableNotFoundError
            AttributeNotFoundError
            ValueError
    """

    field_name = common.AN_SamplingEndTime
    con.verifyAttributeExistence(table_name, field_name)

    field_name = sql.SqlQuery.to_attr_str(field_name)
    select = "DISTINCT MIN(%s),MAX(%s)" % (field_name, field_name)
    result = con.execute_select(select=select, table=table_name, where=where)
    begin_db_time, end_db_time = result.fetchone()
    #import sqlite3
    # print end_db_time, type(end_db_time), sqlite3.dbapi2.converters
    time_format = sql.DB_DATETIME_FORMAT
    datetime_range = gtime.DateTimeRange(
        gtime.toDateTime(begin_db_time, time_format),
        gtime.toDateTime(end_db_time, time_format),
        time_format)
    datetime_range.verifyTimeRange()
    if common.isFloat(discard_rate):
        datetime_range.discard(discard_rate)

    return datetime_range


def getIndexAttrList():
    return [
        common.AN_SamplingStartTime,
        common.AN_SamplingEndTime,
        FN_Cpu,

        "IFACE",

        FN_DiskInformationId,
        FN_DiskDevice,
        FN_DiskStats,
        FN_DiskLdev,
        FN_DiskHdlm,
        FN_DiskLun,
        FN_FileSystem,
    ]


def getNotStatisticalSarDataFieldList():
    field_list = [
        common.AN_GeneralKey,
        # FN_KeyPriority,
        # FN_SysstatFieldOrder,
        # FN_GroupId,
        # FN_GroupPriority,
        FN_Table_Name,
    ]

    pidstat_field = [
        FN_PID,
        FN_COMMAND,

        FN_IrqNumber,
    ]

    return field_list + getIndexAttrList() + pidstat_field


def getStatisticalDataFieldList(con, table_name):
    not_datafield_list = getNotStatisticalSarDataFieldList()

    def isDataAttr(attr):
        return attr not in not_datafield_list

    return filter(isDataAttr, con.getAttributeNameList(table_name))


def __getStats(attr, input_data_list):
    import golib.statistics as stats

    sysstat_order = 0  # del
    ammount_rate = 0

    dict_stats = stats.Calculator.getStatistics(input_data_list)
    sysstat_group = SysstatParser.getSysstatGroupByFieldName(attr)
    meta_info_list = [
        attr,
        0,
        sysstat_group.group_name,
        sysstat_group.group_index,
    ]

    return (
        meta_info_list +
        [dict_stats[stats_key] for stats_key in stats.DATABASE_STATISTICS_LIST] +
        [sysstat_order, ammount_rate]
    )


def getSysstatDataDev(con, parse_option):
    """
    raise:
            TypeError
            TableNotFoundError
            AttributeNotFoundError
            ValueError
    """

    key_attr_list = parse_option.getPrimaryKeyAttrList(con)
    #data_field_list    = getDataFieldList(con, parse_option)
    data_field_list = getStatisticalDataFieldList(
        con, parse_option.table_name)
    if common.isEmptyList(data_field_list):
        return [None, None]

    dict_Field_AllValueList = {}
    value_matrix = []

    datetime_range = getDateTimeRangeFromTableDev(
        con, parse_option.table_name, discard_rate=parse_option.discard_rate)
    where_time_range = " AND ".join(sql.SqlQuery.makeDatetimeRangeWhereList(
        datetime_range, common.AN_SamplingEndTime, common.AN_SamplingEndTime))

    dict_debug = {
        "table": parse_option.table_name,
        "primary key": str(key_attr_list),
        "data field": str(data_field_list),
        "time range": where_time_range,
    }
    logger.debug("sysstat data: " + common.dumpDict(dict_debug))

    if len(key_attr_list) > 0:
        where_list = [
            sql.SqlQuery.make_where(key_attr, "NONE", "!=")
            for key_attr in key_attr_list
        ]
        result = con.execute_select(
            select="DISTINCT %s" % (
                ",".join(sql.SqlQuery.to_attr_str_list(key_attr_list))),
            table=parse_option.table_name,
            where=" AND ".join(where_list))
        if result is None:
            return None
        key_value_matrix = result.fetchall()
        for key_value_list in key_value_matrix:
            where_list = [where_time_range] + [
                sql.SqlQuery.make_where(key_attr, key_value)
                for key_attr, key_value in zip(key_attr_list, key_value_list)
            ]
            result = con.execute_select(
                select=",".join(
                    sql.SqlQuery.to_attr_str_list(data_field_list)),
                table=parse_option.table_name,
                where=" AND ".join(where_list))

            for field, value_list in zip(data_field_list, zip(*result.fetchall())):
                output_data_list = list(
                    key_value_list) + __getStats(field, value_list)
                value_matrix.append(output_data_list)

                if key_value_list[0] not in [common.KEY_AVERAGE, common.KEY_TOTAL]:
                    dict_Field_AllValueList.setdefault(
                        field, []).extend(value_list)

        # key-all data ---
        for field, value_list in six.iteritems(dict_Field_AllValueList):
            output_data_list = (
                [common.KEY_ALL] * len(key_value_list) +
                __getStats(field, value_list)
            )
            value_matrix.append(output_data_list)
    else:
        for field in data_field_list:
            value_list = con.getValueList(
                select=sql.SqlQuery.to_attr_str(field),
                table=parse_option.table_name,
                where=where_time_range)

            output_data_list = __getStats(field, value_list)
            value_matrix.append(output_data_list)

    return (data_field_list, value_matrix)
