# encoding: utf-8

'''
@author: vmmhypervisor
'''

import logging
import re

import six

import golib.common as common
import golib.data_property as dataprop
from golib.logger import logger


class BaseTextWriter(common.BaseObject):

    def __init__(self):
        import sys

        super(BaseTextWriter, self).__init__()

        self.column_separator = "  "
        self.side = ""
        self.is_show_header_separator = True

        self.indent = ""
        self.is_padding = True
        self.min_padding_len = 1

        self.is_numbering = False
        self.sort_key_list = None
        self.is_reverse_sort = False

        self.stream = sys.stdout

        self.__sort_index_list = None

        self._set_indent_level(0)

    def write(self, text):
        raise NotImplementedError()

    def writeLine(self, text):
        raise NotImplementedError()

    def printListList(self, header_list, value_listlist):
        raise NotImplementedError()

    def writeRaw(self, unicode_text):
        self._verify_stream()
        self.stream.write(unicode_text.encode("utf-8", "ignore"))
        # self.stream.write(text)

        return True

    def writeRawLine(self, unicode_text):
        if self.stream is None:
            logger.error("null stream")
            return False

        return self.writeRaw(unicode_text + six.u("\n"))

    def writeChapter(self, text):
        raise NotImplementedError()

    def writeBeginTagLine(self, tag):
        return True

    def writeEndTagLine(self, tag):
        return True

    def _get_indent_string(self):
        return self.indent * self.__indent_level

    def _set_indent_level(self, indent_level):
        self.__indent_level = indent_level

    def _verify_stream(self):
        if self.stream is None:
            raise RuntimeError("null output stream")

    def _verify_input(self, header_list, value_matrix):
        if common.isEmptyListOrTuple(value_matrix):
            logger.debug("empty value list list")
            return

        if common.isEmptyListOrTuple(header_list):
            return

        for row, value_list in enumerate(value_matrix):
            dict_invalid = {}

            if common.isEmptyListOrTuple(value_list):
                dict_invalid["row=%d" % (row)]	= \
                    "not-list=%s" % (str(value_list))
                continue

            if len(header_list) != len(value_list):
                dict_invalid["row=%d" % (row)]	= \
                    "col-size=%d %s" % (len(value_list), str(value_list))

        if len(dict_invalid) > 0:
            import os

            message = [
                "verifyInput: miss match length with header and value",
                "  header: col-size=%d %s" % (
                    len(header_list), header_list),
                "  value:  %s" % (common.dumpDict(dict_invalid)),
            ]
            raise ValueError(os.linesep.join(message))


class TextWriter(BaseTextWriter):

    def _set_up(self):
        pass

    def write(self, text):
        return self.writeRaw(self._get_indent_string() + text)

    def writeLine(self, text=""):
        message = text
        if common.isNotEmptyString(text):
            message = "%s%s%s" % (self.side, text, self.side)

        return self.writeRawLine(self._get_indent_string() + message)

    def writeChapter(self, text):
        self.__indent_level = 0
        return self.writeRawLine("# %s" % (text))

    def writeSection(self, text):
        self.__indent_level = 1
        return_value = self.writeRawLine(text)
        return return_value

    def writeSubSection(self, text):
        self.__indent_level = 2
        return_value = self.writeRawLine(text)
        return return_value

    def writeSubSubSection(self, text):
        self.__indent_level = 3
        return_value = self.writeRawLine(text)
        return return_value


class NullWriter(BaseTextWriter):

    def __init__(self):
        super(NullWriter, self).__init__()

    def writeRaw(self, text):
        pass

    def writeRawLine(self, text):
        pass

    def write(self, text):
        pass

    def writeLine(self, text=""):
        pass

    def printListList(self, header_list, value_listlist):
        pass

    def writeChapter(self, text):
        pass

    def writeSection(self, text):
        pass

    def writeSubSection(self, text):
        pass

    def writeSubSubSection(self, text):
        pass


class ListListWriter(TextWriter):

    def __init__(self):
        super(ListListWriter, self).__init__()

        self.base_indent = "    "
        self.column_separator = "|"
        self.min_padding_len = 3
        self.is_output_last_line = False

    def write(self, text):
        return self.writeRaw(self.indent + text)

    def writeLine(self, text=""):
        message = text
        if common.isNotEmptyString(text):
            message = "%s%s%s" % (self.side, text, self.side)

        return self.writeRawLine(self.indent + message)

    """
    def writeChapter(self, text):
        self.indent = self.base_indent * 0
        return self.writeRawLine("[%s]" % (text))

    def writeSection(self, text):
        self.indent = self.base_indent * 0
        return_value = self.writeRawLine(text)
        self.indent = self.base_indent * 1
        return return_value

    def writeSubSection(self, text):
        self.indent = self.base_indent * 1
        return_value = self.writeRawLine(text)
        self.indent = self.base_indent * 2
        return return_value

    def writeSubSubSection(self, text):
        self.indent = self.base_indent * 2
        return_value = self.writeRawLine(text)
        self.indent = self.base_indent * 3
        return return_value
    #"""

    def printListList(self, header_list, value_matrix):
        self._verify_input(header_list, value_matrix)
        self._set_up()

        # pre-process ---
        self._sortListList(header_list, value_matrix)

        header_list, value_matrix = self._adjustMatrix(
            header_list, value_matrix)

        col_prop_list = dataprop.PropertyExtractor.extractColumnPropertyList(
            header_list, value_matrix)
        value_prop_matrix = dataprop.PropertyExtractor.extractDataPropertyMatrix(
            value_matrix)

        # output header ---
        header_format_list = self._getHeaderFormatListEx(
            header_list, col_prop_list)

        if self._writeHeader(header_format_list, header_list):
            self._writeHeaderSeparator(col_prop_list)

        # output data ---
        value_format_matrix = []
        for value_prop_list in value_prop_matrix:
            value_format_list = []
            for col_prop, value_prop in zip(col_prop_list, value_prop_list):
                work_align = value_prop.align
                formatformat = self._getFormatFormat(work_align)
                if common.isEmptyString(formatformat):
                    logger.error("empty format format")
                    continue

                if value_prop.typecode == dataprop.Typecode.NONE:
                    type_format = value_prop.type_format
                else:
                    type_format = col_prop.type_format

                padding_len = 0
                if self.is_padding:
                    padding_len = max(
                        col_prop.padding_len, self.min_padding_len)

                value_format = formatformat % (padding_len, type_format)
                value_format_list.append(value_format)

            value_format_matrix.append(value_format_list)

        self._writeDataMatrix(value_format_matrix, value_matrix)

    # private method ---
    def _getHeaderFormatListEx(self, header_list, col_prop_list):
        header_format_list = None

        if not common.isEmptyListOrTuple(header_list):
            header_format_list = []
            for col_prop in col_prop_list:
                formatformat = self._getFormatFormat(col_prop.align)
                padding_len = 0
                if self.is_padding:
                    padding_len = max(
                        col_prop.padding_len, self.min_padding_len)
                header_format = formatformat % (padding_len, "s")
                header_format_list.append(header_format)

        return header_format_list

    def _writeHeader(self, header_format_list, header_list):
        if common.isEmptyList(header_format_list):
            if not common.isEmptyList(header_list):
                logger.warn("empty header format")
            return False

        if common.isEmptyList(header_list):
            logger.debug("empty header")
            return False

        if len(header_format_list) != len(header_list):
            message	= """different list size: format and value
				format: %d %s
				value:  %d %s
				""" % (
                len(header_format_list), str(header_format_list),
                len(header_list), str(header_list))

            logger.warn(message)
            return False

        header = self.column_separator.join(
            header_format_list) % tuple(header_list)
        return self.writeLine(header)

    def _writeHeaderSeparator(self, column_prop_list):
        if not self.is_show_header_separator:
            return True

        header_separator_list = []
        for col_prop in column_prop_list:
            padding_len = 0
            if self.is_padding:
                padding_len = max(col_prop.padding_len, self.min_padding_len)
            if not common.isInteger(padding_len):
                continue

            if col_prop.align == dataprop.Align.RIGHT:
                separator_item = "-" * (padding_len - 1) + ":"
            elif col_prop.align == dataprop.Align.CENTER:
                separator_item = ":" + "-" * (padding_len - 2) + ":"
            else:
                separator_item = "-" * padding_len

            header_separator_list.append(separator_item)

        return self.writeLine(
            self.column_separator.join(header_separator_list))

    def _writeDataMatrix(self, format_matrix, data_matrix):
        if common.isEmptyList(format_matrix):
            logger.debug("empty value format")
            return False

        if common.isEmptyList(data_matrix):
            logger.debug("empty value")
            return False

        for format_list, data_list in zip(format_matrix, data_matrix):
            try:
                line = self.column_separator.join(
                    [f % d for f, d in zip(format_list, data_list)])
            except TypeError:
                logger.error(str(format_list) + str(data_list))
            self.writeLine(line)

        # self.writeLine(self.column_separator.join(write_list))

        return True

    def __compareListList(self, rhs_list, lhs_list):
        for sort_index in self.__sort_index_list:
            sort_index = int(sort_index)
            rhs = rhs_list[sort_index]
            lhs = lhs_list[sort_index]
            if cmp(rhs, lhs) != 0:
                # print "!sort: %d, %s %s - %s %s" % (sort_index, lhs,
                # type(lhs), rhs, type(rhs))
                return cmp(rhs, lhs)

        return 0

    def _sortListList(self, header_list, value_matrix):
        if common.isEmptyList(self.sort_key_list):
            return

        logger.debug("sort: key=%s, reverse=%s" % (
            str(self.sort_key_list), self.is_reverse_sort))

        self.__sort_index_list = []
        for sort_key in self.sort_key_list:
            if common.isInteger(sort_key):
                self.__sort_index_list.append(int(sort_key))
                continue

            if common.isString(sort_key):
                if common.isEmptyList(header_list):
                    logger.debug(
                        "could not use sort key (%s): empty header list" % (sort_key))
                    break

                if sort_key in header_list:
                    self.__sort_index_list.append(header_list.index(sort_key))
            else:
                logger.debug("unexpected sort key type: " + str(sort_key))

        if not common.isEmptyList(self.__sort_index_list):
            logger.debug("actual-sort-key=" + str(self.__sort_index_list))

        value_matrix.sort(
            self.__compareListList, reverse=self.is_reverse_sort)
        self.__sort_index_list = None

    def _getLeftAlignFormatFormat(self):
        return "%%-%d%s"

    def _getRightAlignFormatFormat(self):
        return "%%%d%s"

    def _getFormatFormat(self, align):
        """
        DataPeroperty.Align align
        """

        if align == dataprop.Align.AUTO:
            return self._getLeftAlignFormatFormat()
        if align == dataprop.Align.LEFT:
            return self._getLeftAlignFormatFormat()
        if align == dataprop.Align.RIGHT:
            return self._getRightAlignFormatFormat()

        raise ValueError("unknown align: " + str(align))

    def _makeAlignValueFormatEx(self, align, value):
        left_format = "%%-%d%s"
        right_format = "%%%d%s"

        if align == self.ALIGN_LEFT:
            return left_format
        if align == self.ALIGN_RIGHT:
            return right_format

        raise ValueError("unknown align: " + align)

    def _adjustMatrix(self, header_list, value_matrix):
        if common.isEmptyListOrTuple(value_matrix):
            return (header_list, value_matrix)

        if self.is_numbering:
            org_row = len(value_matrix)
            if org_row == 0:
                return (header_list, value_matrix)

            tmp_value_listlist = []
            for row in range(org_row):
                tmp_value_listlist.append([])
                tmp_value_listlist[row]	= \
                    [row + 1] + list(common.getListItem(value_matrix, row))
            value_matrix = tmp_value_listlist

            if header_list is not None:
                header_list = ["no"] + header_list

        # ---
        tmp_value_listlist = []
        for value_list in value_matrix:
            tmp_value_list = []
            for value in value_list:
                value = common.convertValue(value)
                tmp_value_list.append(value)
                """
                if not common.isString(value):
                    tmp_value_list.append(value)
                    continue

                tmp_value_list.append(re.sub("\t", " ", value))
                #"""
            tmp_value_listlist.append(tmp_value_list)
        value_matrix = tmp_value_listlist

        return (header_list, value_matrix)


class JsonWriter(NullWriter):

    def __init__(self):
        super(JsonWriter, self).__init__()

    def printListList(self, header_list, value_matrix):
        self._verify_input(header_list, value_matrix)

        dict_json = {"rows": [
            dict(zip(header_list, value_list)) for value_list in value_matrix
        ]}

        self.stream.write(common.dumpDict(dict_json))


class PukiWikiWriter(ListListWriter):

    def __init__(self):
        super(PukiWikiWriter, self).__init__()
        self.wiki_font_size = 9

    def __writeWikiHeader(self, header_format_list, header_list):
        if header_format_list is None or header_list is None:
            logger.debug("empty header")
            return

        wiki_header_list = []
        for header_format, header in zip(header_format_list, header_list):
            tmp_header = header_format % (header)
            wiki_header_list.append("~" + tmp_header.strip())

        config_list = []
        for header in header_list:
            #config_list.append("RIGHT:SIZE(%d):" % (self.wiki_font_size))
            config_list.append("SIZE(%d):" % (self.wiki_font_size))
        config_line = "%s%s%sc" % (
            self.side, self.column_separator.join(config_list), self.side)
        self.writeLine(config_line)

    def _makeAlignValueFormatEx(self, align, value):
        left_format = "LEFT:%%%d%s"
        right_format = "RIGHT:%%%d%s"

        if align == self.ALIGN_LEFT:
            return left_format
        elif align == self.ALIGN_RIGHT:
            return right_format
        else:
            raise ValueError("unknown align: " + align)

    def _writeHeader(self, header_format_list, header_list):
        if header_format_list is None or header_list is None:
            return

        wiki_header_list = []
        for header_format, header in zip(header_format_list, header_list):
            tmp_header = header_format % (header)
            wiki_header_list.append("~" + tmp_header.strip())
        header_line = "%s%s%s%sh" % (
            self.indent, self.side,
            self.column_separator.join(wiki_header_list), self.side)
        self.writeLine(header_line)
        self.__writeWikiHeader(header_format_list, header_list)

    def _set_up(self):
        self.indent = ""
        self.column_separator = "|"
        self.side = "|"
        self.is_padding = False

    def writeChapter(self, text):
        self.writeRawLine("* %s" % (text))

    def writeSection(self, text):
        self.writeRawLine("** %s" % (text))

    def writeSubSection(self, text):
        self.writeRawLine("*** %s" % (text))

    def writeSubSubSection(self, text):
        self.writeRawLine("- %s" % (text))


class MarkdownWriter(ListListWriter):

    def __init__(self):
        super(MarkdownWriter, self).__init__()
        self.wiki_font_size = 9

    def __writeWikiHeader(self, header_format_list, header_list):
        if header_format_list is None or header_list is None:
            logger.debug("empty header")
            return

        config_list = []
        for _header in header_list:
            config_list.append("----")
        config_line = "%s%s%sc" % (
            self.side, self.column_separator.join(config_list), self.side)
        self.writeLine(config_line)

    def _makeAlignValueFormatEx(self, align, value):
        # tmp
        left_format = "%%%d%s"
        right_format = "%%%d%s"

        if align == self.ALIGN_LEFT:
            return left_format
        elif align == self.ALIGN_RIGHT:
            return right_format
        else:
            raise ValueError("unknown align: " + align)

    def _writeHeader(self, header_format_list, header_list):
        if header_format_list is None or header_list is None:
            return

        wiki_header_list = []
        for header_format, header in zip(header_format_list, header_list):
            tmp_header = header_format % (header)
            wiki_header_list.append(tmp_header.strip())
        header_line = "%s%s%s%s" % (
            self.indent, self.side,
            self.column_separator.join(wiki_header_list), self.side)
        self.writeLine(header_line)
        self.__writeWikiHeader(header_format_list, header_list)

    def _set_up(self):
        self.indent = ""
        self.column_separator = "|"
        #self.side		= "|"
        self.is_padding = False

    def writeChapter(self, text):
        self.writeRawLine("# %s" % (text))

    def writeSection(self, text):
        self.writeRawLine("## %s" % (text))

    def writeSubSection(self, text):
        self.writeRawLine("### %s" % (text))

    def writeSubSubSection(self, text):
        self.writeRawLine("- %s" % (text))


class CsvWriter(ListListWriter):

    def __init__(self):
        super(CsvWriter, self).__init__()

    def _set_up(self):
        self.indent = ""
        self.column_separator = ","
        self.side = ""
        self.is_padding = False
        self.is_output_last_line = False
        self.is_show_header_separator = False

    def writeChapter(self, text):
        pass

    def writeSection(self, text):
        pass

    def writeSubSection(self, text):
        pass

    def writeSubSubSection(self, text):
        pass


class LoggerWriter(ListListWriter):
    RE_LINESEP = re.compile("\n")

    def __init__(self):
        super(LoggerWriter, self).__init__()
        self.log_level = logging.INFO

        self.is_output_last_line = True

    def writeRaw(self, text):
        self._verify_stream()

        caller = logging.getLogger().findCaller()
        logger.write(self.RE_LINESEP.sub("", text), self.log_level, caller)
        return True


debug_writer = LoggerWriter()
debug_writer.log_level = logging.DEBUG
debug_writer.is_output_last_line = True


def debugMatrix(header_list, value_matrix):
    debug_writer.printListList(header_list, value_matrix)


class HtmlWrter(ListListWriter):
    __HTML_INDENT = "  "

    def __init__(self):
        super(HtmlWrter, self).__init__()

        self.vertical_align = "top"
        self.cell_padding = 6
        self.caption = None
        self.style = None

        self.warn_threashold = None
        self.danger_threashold = None

        self.__nest_level = 0

    def _set_up(self):
        self.indent = ""
        self.column_separator = ""
        self.side = ""
        self.is_padding = False

    def __getTagText(self, tag, attribute_list):
        tag = tag.strip()
        tag_list = [tag] + attribute_list
        return "<%s>" % (" ".join(tag_list))

    def __getIndent(self):
        return self.__nest_level * self.__HTML_INDENT

    def incrementIndent(self):
        self.__nest_level += 1

    def decrementIndent(self):
        self.__nest_level -= 1

    def writeComment(self, comment):
        comment_html = "<!-- %s -->" % (comment.strip())
        self.writeRawLine(self.__getIndent() + comment_html)

    def writeBeginTag(self, tag, attribute_list=[]):
        #tag_list	= [tag] + attribute_list
        self.writeRaw(
            self.__getIndent() + self.__getTagText(tag, attribute_list))
        #self.write(indent + "<%s>" % (" ".join(tag_list)))

    def writeBeginTagLine(self, tag, attribute_list=[]):
        """
        tag_list	= [tag] + attribute_list
        self.writeRawLine(indent + "<%s>" % (" ".join(tag_list)))
        """
        self.writeRawLine(
            self.__getIndent() + self.__getTagText(tag, attribute_list))
        self.incrementIndent()

    def writeEndTag(self, tag):
        tag = tag.strip()
        self.writeRawLine("</%s>" % (tag))

    def writeEndTagLine(self, tag):
        self.decrementIndent()
        tag = tag.strip()
        self.writeRawLine(self.__getIndent() + "</%s>" % (tag))

    def writeTag(self, tag, body, attribute_list=[]):
        self.writeBeginTag(tag, attribute_list)
        self.writeRaw(body.strip())
        self.writeEndTag(tag)

    def writeHtmlHead(self, title, css_path):
        self.writeRawLine(
            "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\"\n"
            "\"http://www.w3.org/TR/html4/strict.dtd\">")

        self.writeBeginTagLine("head")

        if common.isNotEmptyString(css_path):
            self.writeRawLine(
                "<link rel='stylesheet' href='%s' type='text/css'>" % (css_path))

        if common.isNotEmptyString(title):
            self.writeBeginTagLine("title")
            self.write("%s\n" % (title))
            self.writeEndTagLine("title")

        self.writeEndTagLine("head")

    def __writeHeader(self, header, text):
        import golib.gfile

        anchor_id = golib.gfile.sanitizeFileName(text)
        self.writeBeginTagLine(header, ["id='%s'" % (anchor_id)])
        self.write(text + "\n")
        self.writeEndTagLine(header)

    def writeChapter(self, text):
        self.__writeHeader("h1", text)

    def writeSection(self, text):
        self.__writeHeader("h2", text)

    def writeSubSection(self, text):
        self.__writeHeader("h3", text)

    def writeSubSubSection(self, text):
        self.__writeHeader("h4", text)

    def write(self, text):
        self.indent = self.__getIndent()
        super(HtmlWrter, self).write(text)

    def writeLine(self, text):
        #self.stream.write("<tr valign='%s'>" % (self.vertical_align) + "\n")

        self.writeBeginTagLine("tr", ["valign='%s'" % (self.vertical_align)])
        #super(HtmlWrter, self).writeLine(self.indent + text)
        self.indent = self.__getIndent()
        super(HtmlWrter, self).writeLine(text)
        self.writeEndTagLine("tr")

    def _writeHeader(self, header_format_list, header_list):
        if common.isEmptyListOrTuple(header_format_list) or common.isEmptyListOrTuple(header_list):
            return

        work_header_list = []
        for header_format, header in zip(header_format_list, header_list):
            tmp_header = header_format % (header)
            #work_header_list.append("<th hspace='%d'>%s</th>" % (self.hspace, tmp_header.strip()))
            work_header_list.append("%s" % (tmp_header.strip()))
            #work_header_list.append("<th>%s</th>" % (tmp_header.strip()))
            # self.writeLine(tmp_header.strip())
            #self.write("<th>%s</th>\n" % (tmp_header.strip()))

        header_line = "%s%s%s%s" % (
            self.indent, self.side,
            self.column_separator.join(work_header_list), self.side)
        self.writeLine(header_line)

    def _makeAlignValueFormatEx(self, align, value):
        font_tag_head = ""
        font_tag_tail = ""
        font_color = None

        if common.isFloat(value):
            if value == 0:
                font_color = "#888888"
            else:
                if self.warn_threashold is not None:
                    if value >= self.warn_threashold:
                        font_color = "darkorange"

                if self.danger_threashold is not None:
                    if value >= self.danger_threashold:
                        font_color = "red"
        elif value is None or common.isNaN(value):
            font_color = "#cccccc"

        if common.isNotEmptyString(font_color):
            font_tag_head = "<font color='%s'>" % (font_color)
            font_tag_tail = "</font>"

        left_format = "<td align='left'>" + font_tag_head + \
            "%%%d%s" + font_tag_tail + "</td>"
        right_format = "<td align='right'>" + \
            font_tag_head + "%%%d%s" + font_tag_tail + "</td>"

        if align == self.ALIGN_LEFT:
            return left_format
        elif align == self.ALIGN_RIGHT:
            return right_format

        raise ValueError("unknown align: " + align)

    def _getLeftAlignFormatFormat(self):
        return "<td align='left'>" + "%%%d%s" + "</td>"

    def _getRightAlignFormatFormat(self):
        return "<td align='right'>" + "%%%d%s" + "</td>"

    def printListList(self, header_list, value_listlist):
        attr_list = [
            "cellpadding='%d'" % (self.cell_padding),
            "border='1'",
        ]
        if common.isNotEmptyString(self.style):
            attr_list.append("class='%s'" % (self.style))
        self.writeBeginTagLine("table", attr_list)
        if common.isNotEmptyString(self.caption):
            self.writeTag("caption", self.caption)

        super(HtmlWrter, self).printListList(header_list, value_listlist)
        self.writeEndTagLine("table")
