'''
@author: vmmhypervisor

:required:
    https://pypi.python.org/pypi/pytest
'''

import os
import datetime

import pytest

import golib
from golib.sysstat import *
from golib.gfile import *
from golib.gtime import *


TEST_DIR_PATH = os.path.dirname(os.path.realpath(__file__))
TEST_SAR_BIN_PATH = os.path.join(TEST_DIR_PATH, "sar.dat")

EMPTY_FILE_NAME = "empty_file"
TEMP_FILE_NAME = "tmp_file"


@pytest.fixture
def sys_wrapper():
    return golib.syswrapper.SystemWrapper()


class Test_verifySarBinDataFile:

    @pytest.mark.parametrize(["value"], [
        [TEST_SAR_BIN_PATH],
    ])
    def test_normal(self, sys_wrapper, value):
        Sar.initialize(sys_wrapper)
        Sar.verifyBinDataFile(value)

    def test_abnormal(self, tmpdir):
        p = tmpdir.join(TEMP_FILE_NAME)
        p.write("test")

        test_path = str(tmpdir.join(TEMP_FILE_NAME))

        with pytest.raises(IOError):
            Sar.verifyBinDataFile(test_path)

    @pytest.mark.parametrize(["value", "expected"], [
        [None, TypeError],
        ["__not_existing_path__", FileNotFoundError],
    ])
    def test_exception(self, sys_wrapper, value, expected):
        Sar.initialize(sys_wrapper)
        with pytest.raises(expected):
            Sar.verifyBinDataFile(value)

    def test_exception_empty_file(self, sys_wrapper, tmpdir):
        p = tmpdir.join(EMPTY_FILE_NAME)
        p.write("")

        Sar.initialize(sys_wrapper)
        with pytest.raises(EmptyFileError):
            Sar.verifyBinDataFile(str(tmpdir.join(EMPTY_FILE_NAME)))


class Test_isValidBinDataFile:

    @pytest.mark.parametrize(["value", "expected"], [
        [TEST_SAR_BIN_PATH, True],
    ])
    def test_normal(self, sys_wrapper, value, expected):
        Sar.initialize(sys_wrapper)
        assert Sar.isValidBinDataFile(value) == expected

    def test_abnormal(self, sys_wrapper, tmpdir):
        Sar.initialize(sys_wrapper)

        p = tmpdir.join(TEMP_FILE_NAME)
        p.write("test")

        test_path = str(tmpdir.join(TEMP_FILE_NAME))

        assert Sar.isValidBinDataFile(test_path) == False

    @pytest.mark.parametrize(["value", "expected"], [
        [None, TypeError],
        ["__not_existing_path__", FileNotFoundError],
    ])
    def test_exception(self, sys_wrapper, value, expected):
        Sar.initialize(sys_wrapper)
        with pytest.raises(expected):
            Sar.isValidBinDataFile(value)

    def test_exception_empty_file(self, sys_wrapper, tmpdir):
        p = tmpdir.join(EMPTY_FILE_NAME)
        p.write("")

        Sar.initialize(sys_wrapper)
        with pytest.raises(EmptyFileError):
            Sar.isValidBinDataFile(str(tmpdir.join(EMPTY_FILE_NAME)))


class Test_verifyTextDataFile:
    NORMAL_TEST_DATA	= """Linux 3.2.0-4-686-pae (debian7)         04/11/15        _i686_
21:14:42        CPU     %user     %nice   %system   %iowait    %steal     %idle
21:14:43        all      2.00      0.00      3.00      2.00      0.00     93.00
21:14:44        all      0.50      0.00      0.50      0.50      0.00     98.51
21:14:45        all      0.00      0.00      0.50      1.01      0.00     98.49
21:14:46        all      0.50      0.00      1.00      0.50      0.00     98.00
21:14:47        all      0.00      0.00      0.00      1.01      0.00     98.99
"""

    ERROR_TEST_DATA_1 = "hoge"
    ERROR_TEST_DATA_2	= """
hoge
hoge
hoge
hoge
"""

    def test_normal(self, tmpdir):
        p = tmpdir.join(TEMP_FILE_NAME)
        p.write(self.NORMAL_TEST_DATA)

        test_path = str(tmpdir.join(TEMP_FILE_NAME))
        Sar.verifyTextDataFile(test_path)

    @pytest.mark.parametrize(["value", "expected"], [
        [ERROR_TEST_DATA_1, IOError],
        [ERROR_TEST_DATA_2, ValueError],
    ])
    def test_abnormal(self, value, expected, tmpdir):
        p = tmpdir.join(TEMP_FILE_NAME)
        p.write(value)

        test_path = str(tmpdir.join(TEMP_FILE_NAME))
        with pytest.raises(expected):
            Sar.verifyTextDataFile(test_path)

    @pytest.mark.parametrize(["value", "expected"], [
        [TEST_SAR_BIN_PATH, IOError],
        [None, TypeError],
        ["__not_existing_path__", FileNotFoundError],
    ])
    def test_exception(self, sys_wrapper, value, expected):
        Sar.initialize(sys_wrapper)
        with pytest.raises(expected):
            Sar.verifyTextDataFile(value)


class Test_parseSysstatHeader:

    @pytest.mark.parametrize(["value", "expected"], [
        [
            "Linux 3.2.0-4-686-pae (debian7)         04/11/15        _i686_",
            ["debian7", datetime.datetime.strptime(
                "2015/04/11", Format.JST_DATE).date()],
        ],
    ])
    def test_normal(self, value, expected):
        hostname, sysstat_datetim = SysstatParser.parseSysstatHeader(value)
        assert hostname == expected[0]
        assert sysstat_datetim == expected[1]

    @pytest.mark.parametrize(["value", "expected"], [
        [None, TypeError],
        [1, TypeError],
        [True, TypeError],
        ["aaa bbb", ValueError],
        ["aaa bbb ccc ddd", ValueError],
    ])
    def test_exception(self, sys_wrapper, value, expected):
        with pytest.raises(expected):
            SysstatParser.parseSysstatHeader(value)
